from math import sqrt
import scipy.sparse.linalg as sla
from matplotlib import pyplot
import kwant

syst = kwant.Builder()
lat = kwant.lattice.general([(1,0,0), (0,1,0), (0,0,1)])

t = 1.0
R = 10

syst[(lat(x,y,z) for x in range(-R-1, R+1)
     for y in range(-R-1, R+1) for z in range(R+1)
     if sqrt(x**2 + y**2 + z**2) < R + 0.01)] = 4 * t
syst[(lat(x,y,z) for x in range(-2*R, 2*R + 1)
     for y in range(-R, R+1) for z in range(-R, 0))] = 4 * t
syst[lat.neighbors()] = -t
syst = syst.finalized()
kwant.plot(syst)

ham = syst.hamiltonian_submatrix(sparse=True)
wf = abs(sla.eigs(ham, sigma=0.2, k=6)[1][:,0])**2
size = 3 * wf / wf.max()
kwant.plot(syst, site_size=size, site_color=(0,0,1,0.05), site_lw=0)



#
# (2018) Modified by Kwant Authors
#
# Modifications
# =============
# Merged and modified from scipy/sparse/csgraph/_shortest_path.pyx
#
# All shortest path algorithms except for Dijkstra's removed.
# Implementation of Dijkstra's algorithm modified to allow for specific
# use-cases required by Flux. The changes are documented in the docstring.
#
#
# Copyright (c) 2001, 2002 Enthought, Inc.
# All rights reserved.
#
# Copyright (c) 2003-2017 SciPy Developers.
# All rights reserved.
#
# Copyright (c) 2011 Jake Vanderplas <vanderplas@astro.washington.edu>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of Enthought nor the names of the SciPy Developers
#      may be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
cimport numpy as np
from numpy.math cimport INFINITY as inf

from libc.stdlib cimport malloc, free
from libc.string cimport memset

ctypedef np.float64_t DTYPE_t
ctypedef np.int32_t ITYPE_t
ITYPE = np.int32


def dijkstra_directed(
        object graph,
        ITYPE_t[:] sources,
        ITYPE_t[:] targets,
        bint return_paths=True,
        bint return_predecessors=False):
    """Modified directed Dijkstra algorithm.

    Edges with infinite weight are treated as if they do not exist.

    The shortest paths between edge pairs 'zip(sources, targets)'
    are found.

    If 'len(sources) == 1' then this routine can be used to
    compute the one-to-all distances by passing an integer
    greater than any node in 'graph' as the 'target'.
    In this case 'return_predecessors' may be specified, and
    the predecessor matrix will also be returned.

    'return_predecessors' and 'return_paths' are mutually exclusive.

    Returns
    -------
    if return_paths:
        (paths, path_lengths)
    elif return_predecessors:
        (path_lengths, predecessors)
    else:
        path_lengths

    """
# Implementation of Dijkstra's algorithm modified to allow for
# early exit (when target is found) and return the path from source
# to target, rather than the whole predecessor matrix. In addition
# graph edges with infinite weight are treated as if they do not exist.

    cdef ITYPE_t[:] csr_indices = graph.indices, csr_indptr = graph.indptr
    cdef DTYPE_t[:] csr_weights = graph.data

    # This implementation of Dijkstra's algorithm is very tightly coupled
    # to our use-case in 'flux', so we allow ourselves to assert
    assert sources.shape[0] == targets.shape[0]
    assert graph.shape[0] == graph.shape[1]
    assert not (return_predecessors and return_paths)
    assert not (return_predecessors and (sources.shape[0] > 1))

    cdef unsigned int num_links = sources.shape[0], num_nodes = graph.shape[0]

    cdef unsigned int i, k, j_source, j_target, j_current
    cdef ITYPE_t j

    cdef DTYPE_t next_val

    cdef FibonacciHeap heap
    cdef FibonacciNode *v
    cdef FibonacciNode *current_node
    cdef FibonacciNode* nodes = <FibonacciNode*> malloc(num_nodes * sizeof(FibonacciNode))

    cdef ITYPE_t[:] pred = np.empty((num_nodes,), dtype=ITYPE)


    # outputs
    cdef DTYPE_t[:] path_lengths = np.zeros((num_links,), float)
    cdef list paths
    if return_paths:
        paths = []

    for i in range(num_links):
        j_source = sources[i]
        j_target = targets[i]

        for k in range(num_nodes):
            initialize_node(&nodes[k], k)
        pred[:] = -1  # only useful for debugging

        heap.min_node = NULL
        insert_node(&heap, &nodes[j_source])

        while heap.min_node:
            v = remove_min(&heap)
            v.state = SCANNED

            if v.index == j_target:
                path_lengths[i] = v.val
                if return_paths:
                    paths.append(_calculate_path(pred, j_source, j_target))
                break  # next iteration of outer 'for' loop

            for j in range(csr_indptr[v.index], csr_indptr[v.index + 1]):
                if csr_weights[j] == inf:
                    # Treat infinite weight links as missing
                    continue
                j_current = csr_indices[j]
                current_node = &nodes[j_current]
                if current_node.state != SCANNED:
                    next_val = v.val + csr_weights[j]
                    if current_node.state == NOT_IN_HEAP:
                        current_node.state = IN_HEAP
                        current_node.val = next_val
                        insert_node(&heap, current_node)
                        pred[j_current] = v.index
                    elif current_node.val > next_val:
                        decrease_val(&heap, current_node,
                                     next_val)
                        pred[j_current] = v.index

    free(nodes)
    if return_paths:
      return paths, path_lengths
    elif return_predecessors:
      return path_lengths, pred
    else:
        return path_lengths


cdef list _calculate_path(ITYPE_t[:] pred, int j_source, int j_target):
    visited = []
    cdef int node = j_target
    while node != j_source:
        visited.append(node)
        node = pred[node]
    visited.append(j_source)
    return visited


######################################################################
# FibonacciNode structure
#  This structure and the operations on it are the nodes of the
#  Fibonacci heap.
#
cdef enum FibonacciState:
    SCANNED
    NOT_IN_HEAP
    IN_HEAP


cdef struct FibonacciNode:
    unsigned int index
    unsigned int rank
    FibonacciState state
    DTYPE_t val
    FibonacciNode* parent
    FibonacciNode* left_sibling
    FibonacciNode* right_sibling
    FibonacciNode* children


cdef void initialize_node(FibonacciNode* node,
                          unsigned int index,
                          DTYPE_t val=0):
    # Assumptions: - node is a valid pointer
    #              - node is not currently part of a heap
    node.index = index
    node.val = val
    node.rank = 0
    node.state = NOT_IN_HEAP

    node.parent = NULL
    node.left_sibling = NULL
    node.right_sibling = NULL
    node.children = NULL


cdef FibonacciNode* rightmost_sibling(FibonacciNode* node):
    # Assumptions: - node is a valid pointer
    cdef FibonacciNode* temp = node
    while(temp.right_sibling):
        temp = temp.right_sibling
    return temp


cdef FibonacciNode* leftmost_sibling(FibonacciNode* node):
    # Assumptions: - node is a valid pointer
    cdef FibonacciNode* temp = node
    while(temp.left_sibling):
        temp = temp.left_sibling
    return temp


cdef void add_child(FibonacciNode* node, FibonacciNode* new_child):
    # Assumptions: - node is a valid pointer
    #              - new_child is a valid pointer
    #              - new_child is not the sibling or child of another node
    new_child.parent = node

    if node.children:
        add_sibling(node.children, new_child)
    else:
        node.children = new_child
        new_child.right_sibling = NULL
        new_child.left_sibling = NULL
        node.rank = 1


cdef void add_sibling(FibonacciNode* node, FibonacciNode* new_sibling):
    # Assumptions: - node is a valid pointer
    #              - new_sibling is a valid pointer
    #              - new_sibling is not the child or sibling of another node
    cdef FibonacciNode* temp = rightmost_sibling(node)
    temp.right_sibling = new_sibling
    new_sibling.left_sibling = temp
    new_sibling.right_sibling = NULL
    new_sibling.parent = node.parent
    if new_sibling.parent:
        new_sibling.parent.rank += 1


cdef void remove(FibonacciNode* node):
    # Assumptions: - node is a valid pointer
    if node.parent:
        node.parent.rank -= 1
        if node.left_sibling:
            node.parent.children = node.left_sibling
        elif node.right_sibling:
            node.parent.children = node.right_sibling
        else:
            node.parent.children = NULL

    if node.left_sibling:
        node.left_sibling.right_sibling = node.right_sibling
    if node.right_sibling:
        node.right_sibling.left_sibling = node.left_sibling

    node.left_sibling = NULL
    node.right_sibling = NULL
    node.parent = NULL


######################################################################
# FibonacciHeap structure
#  This structure and operations on it use the FibonacciNode
#  routines to implement a Fibonacci heap

ctypedef FibonacciNode* pFibonacciNode


cdef struct FibonacciHeap:
    FibonacciNode* min_node
    pFibonacciNode[100] roots_by_rank  # maximum number of nodes is ~2^100.


cdef void insert_node(FibonacciHeap* heap,
                      FibonacciNode* node):
    # Assumptions: - heap is a valid pointer
    #              - node is a valid pointer
    #              - node is not the child or sibling of another node
    if heap.min_node:
        add_sibling(heap.min_node, node)
        if node.val < heap.min_node.val:
            heap.min_node = node
    else:
        heap.min_node = node


cdef void decrease_val(FibonacciHeap* heap,
                       FibonacciNode* node,
                       DTYPE_t newval):
    # Assumptions: - heap is a valid pointer
    #              - newval <= node.val
    #              - node is a valid pointer
    #              - node is not the child or sibling of another node
    #              - node is in the heap
    node.val = newval
    if node.parent and (node.parent.val >= newval):
        remove(node)
        insert_node(heap, node)
    elif heap.min_node.val > node.val:
        heap.min_node = node


cdef void link(FibonacciHeap* heap, FibonacciNode* node):
    # Assumptions: - heap is a valid pointer
    #              - node is a valid pointer
    #              - node is already within heap

    cdef FibonacciNode *linknode
    cdef FibonacciNode *parent
    cdef FibonacciNode *child

    if heap.roots_by_rank[node.rank] == NULL:
        heap.roots_by_rank[node.rank] = node
    else:
        linknode = heap.roots_by_rank[node.rank]
        heap.roots_by_rank[node.rank] = NULL

        if node.val < linknode.val or node == heap.min_node:
            remove(linknode)
            add_child(node, linknode)
            link(heap, node)
        else:
            remove(node)
            add_child(linknode, node)
            link(heap, linknode)


cdef FibonacciNode* remove_min(FibonacciHeap* heap):
    # Assumptions: - heap is a valid pointer
    #              - heap.min_node is a valid pointer
    cdef FibonacciNode *temp
    cdef FibonacciNode *temp_right
    cdef FibonacciNode *out
    cdef unsigned int i

    # make all min_node children into root nodes
    if heap.min_node.children:
        temp = leftmost_sibling(heap.min_node.children)
        temp_right = NULL

        while temp:
            temp_right = temp.right_sibling
            remove(temp)
            add_sibling(heap.min_node, temp)
            temp = temp_right

        heap.min_node.children = NULL

    # choose a root node other than min_node
    temp = leftmost_sibling(heap.min_node)
    if temp == heap.min_node:
        if heap.min_node.right_sibling:
            temp = heap.min_node.right_sibling
        else:
            out = heap.min_node
            heap.min_node = NULL
            return out

    # remove min_node, and point heap to the new min
    out = heap.min_node
    remove(heap.min_node)
    heap.min_node = temp

    # re-link the heap
    for i in range(100):
        heap.roots_by_rank[i] = NULL

    while temp:
        if temp.val < heap.min_node.val:
            heap.min_node = temp
        temp_right = temp.right_sibling
        link(heap, temp)
        temp = temp_right

    return out



"""An example of advanced system creation."""

from math import tanh
from cmath import exp
import tinyarray as ta
import kwant


sigma_0 = ta.identity(2)
sigma_x = ta.array([[0, 1], [1, 0]])
sigma_y = ta.array([[0, -1j], [1j, 0]])
sigma_z = ta.array([[1, 0], [0, -1]])


def make_system(R):
    def in_ring(pos):
        x, y = pos
        return R**2 / 4 < x**2 + y**2 < R**2

    def in_lead(pos):
        x, y = pos
        return -R / 4 < y < R / 4

    def pot(site, B):
        x, y = site.pos
        return (0.1 * tanh(x / R) + tanh(2 * y / R)) * sigma_z

    def hop(site1, site2, B):
        x1, y1 = site1.pos
        x2, y2 = site2.pos
        return - exp(.5j * B * (x1 - x2) * (y1 + y2)) * sigma_0

    lat = kwant.lattice.honeycomb()

    syst = kwant.Builder()
    syst[lat.shape(in_ring, (3 * R / 4, 0))] = pot
    syst[lat.neighbors()] = hop

    lead = kwant.Builder(kwant.TranslationalSymmetry(lat.vec((-1, 0))))
    lead[lat.shape(in_lead, (0, 0))] = sigma_0
    lead[lat.neighbors()] = sigma_x
    syst.attach_lead(lead)
    syst.attach_lead(lead.reversed())

    return syst.finalized()


def main():
    syst = make_system(100)
    print(kwant.smatrix(syst, 1.1, params=dict(B=0.1)).transmission(0, 1))


if __name__ == '__main__':
    main()



# Copyright 2011-2016 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

"""Directed graphs optimized for storage and runtime efficiency."""

__all__ = ['Graph', 'CGraph']

# In this module we manage arrays directly with malloc, realloc and free to
# circumvent two problems with Cython:
#
# (1) There are no efficient arrays which allow appending (numpy.ndarray
#     doesn't).
#
# (2) Cython extension types cannot have typed buffers as members.
#
# Once these two problems are solved, the corresponding code could be
# rewritten, probably to use Python's array.array.

# TODO: represent all dangling nodes by -1

# TODO (perhaps): transform Graph into something which behaves like a python
# sequence.  Allow creation of compressed graphs from any sequence.

from libc.stdlib cimport malloc, realloc, free
from libc.string cimport memset
from cpython cimport array
import array
import numpy as np
cimport numpy as np
from .defs cimport gint

cdef class Graph:
    """An uncompressed graph.  Used to make compressed graphs.  (See `CGraph`.)
    """
    # The array edges holds `size` elements with space for `capacity`.

    def __init__(self, allow_negative_nodes=False):
        self.allow_negative_nodes = allow_negative_nodes

    def __dealloc__(self):
        free(self.edges)

    property num_nodes:
        def __get__(self):
            return self._num_nodes

        def __set__(self, value):
            if value < self._num_nodes:
                raise ValueError("The number of nodes cannot be decreased.")
            self._num_nodes = value

    cpdef reserve(self, gint capacity):
        """Reserve space for edges.

        Parameters
        ----------
        capacity : integer
           Number of edges for which to reserve space.

        Notes
        -----
        It is not necessary to call this method, but using it can speed up the
        creation of graphs.
        """
        if capacity <= self.capacity:
            return
        self.edges = <Edge*>realloc(self.edges, capacity * sizeof(Edge))
        if not self.edges:
            raise MemoryError
        self.capacity = capacity

    cpdef gint add_edge(self, gint tail, gint head) except -1:
        """Add the directed edge (`tail`, `head`) to the graph.

        Parameters
        ----------
        tail : integer
        head : integer

        Raises
        ------
        ValueError
            If a negative node is added when this has not been allowed
            explicitly or if an edge is doubly-dangling.

        Returns
        -------
        edge_nr : integer
           The sequential number of the edge.  This number can be used to query
           for the edge ID of an edge in the compressed graph.
        """
        cdef bint neg_tail = tail < 0
        cdef bint neg_head = head < 0
        if neg_tail or neg_head:
            if not self.allow_negative_nodes:
                raise ValueError(
                    "Negative node numbers have to be allowed explicitly.")
            if neg_tail and neg_head:
                raise ValueError("Doubly-dangling edges are never allowed.")
            if neg_head:
                self.num_pn_edges += 1
            else:
                self.num_np_edges += 1
        else:
            self.num_pp_edges += 1

        if self.size == self.capacity:
            if self.capacity == 0:
                self.reserve(8)
            else:
                self.reserve(2 * self.capacity)
        self.edges[self.size].tail = tail
        self.edges[self.size].head = head
        self.size += 1
        self._num_nodes = max(self._num_nodes, tail + 1, head + 1)
        return self.size - 1

    def add_edges(self, edges):
        """Add multiple edges in one pass.

        Parameters
        ----------
        edges : iterable of 2-sequences of integers
            The parameter `edges` must be an iterable of elements which
            describe the edges to be added.  For each edge-element, edge[0] and
            edge[1] must give, respectively, the tail and the head.  Valid
            edges are, for example, a list of 2-integer-tuples, or an
            numpy.ndarray of integers with a shape (n, 2).  The latter case is
            optimized.

        Returns
        -------
        first_edge_nr : integer
           The sequential number of the first of the added edges.  The numbers
           of the other added edges are consecutive integers following the
           number of the first.  Edge numbers can be used to query for the edge
           ID of an edge in the compressed graph.
        """
        result = self.size
        if isinstance(edges, np.ndarray):
            if edges.dtype == np.int64:
                self._add_edges_ndarray_int64(edges)
            elif edges.dtype == np.int32:
                self._add_edges_ndarray_int32(edges)
            else:
                self._add_edges_ndarray_int64(edges.astype(np.int64))
        else:
            for edge in edges:
                self.add_edge(*edge)
        return result

    cdef _add_edges_ndarray_int64(self, np.ndarray[np.int64_t, ndim=2] edges):
        cdef int i
        for i in range(edges.shape[0]):
            self.add_edge(edges[i, 0], edges[i, 1])

    cdef _add_edges_ndarray_int32(self, np.ndarray[np.int32_t, ndim=2] edges):
        cdef int i
        for i in range(edges.shape[0]):
            self.add_edge(edges[i, 0], edges[i, 1])

    def compressed(self, bint twoway=False, bint edge_nr_translation=False,
                   bint allow_lost_edges=False):
        """Build a CGraph from this graph.

        Parameters
        ----------
        twoway : boolean (default: False)
            If set, it will be possible to query the compressed graph for
            incoming neighbors.
        edge_nr_translation : boolean (default: False)
            If set, it will be possible to call the method `edge_id`.
        allow_lost_edges : boolean (default: False)
            If set, negative tails are accepted even with one-way compression.

        Raises
        ------
        ValueError
            When negative tails occur while `twoway` and `allow_lost_edges` are
            both false.

        Notes
        -----
        In a one-way compressed graph, an edge with a negative tail is present
        only minimally: it is only possible to query the head of such an edge,
        given the edge ID.  This is why one-way compression of a graph with a
        negative tail leads to a ValueError being raised, unless
        `allow_lost_edges` is true.
        """
        assert (self.size ==
                self.num_pp_edges + self.num_pn_edges + self.num_np_edges)
        if not (twoway or allow_lost_edges or self.num_np_edges == 0):
            raise ValueError('Edges with negative tails cannot be '
                             'represented in an one-way compressed graph.')

        cdef gint s, tail, head, edge_nr
        cdef CGraph_malloc result = CGraph_malloc(twoway, edge_nr_translation,
                                                  self._num_nodes,
                                                  self.num_pp_edges,
                                                  self.num_pn_edges,
                                                  self.num_np_edges)
        cdef gint *hbuf = result.heads_idxs + 1
        cdef gint *heads = result.heads
        cdef gint *tbuf = result.tails_idxs + 1
        cdef gint *tails = result.tails
        cdef gint *edge_ids = result.edge_ids
        cdef gint edge_id = 0, num_edges     # = 0 is there to silence warning.
        cdef Edge *edge

        # `hbuf` is just `heads_idxs` shifted by one.  We will use `hbuf` to
        # build up `heads` and its end state will be such that `heads_idxs`
        # will have the right content.  For `tbuf`, replace "head" with tail in
        # the previous text.

        # Make a histogram of outgoing edges per node in `hbuf` and one of
        # incoming edges per node in `tbuf`.
        memset(result.heads_idxs, 0, (self._num_nodes + 1) * sizeof(gint))
        if twoway:
            memset(result.tails_idxs, 0, (self._num_nodes + 1) * sizeof(gint))
        for edge in self.edges[:self.size]:
            if edge.tail >= 0:
                hbuf[edge.tail] += 1
            if twoway and edge.head >= 0:
                tbuf[edge.head] += 1

        # Replace `hbuf` with its "antiderivative" and then subtract the
        # original `hbuf` from it.  This is done in one pass.
        s = 0
        for tail in range(self._num_nodes):
            s += hbuf[tail]
            hbuf[tail] = s - hbuf[tail]

        # Same as before for `tbuf`.
        if twoway:
            s = 0
            for head in range(self._num_nodes):
                s += tbuf[head]
                tbuf[head] = s - tbuf[head]

        # Iterate through all edges and build `heads` and `tails`.
        next_np_edge_id = result.num_px_edges
        edge_nr = 0
        for edge in self.edges[:self.size]:
            if edge.tail >= 0:
                edge_id = hbuf[edge.tail]
                hbuf[edge.tail] += 1
            elif twoway:
                assert edge.head >= 0
                edge_id = next_np_edge_id
                next_np_edge_id += 1
            else:
                edge_id = -1
            if edge_id >= 0:
                heads[edge_id] = edge.head
            if twoway and edge.head >= 0:
                tails[tbuf[edge.head]] = edge.tail
                edge_ids[tbuf[edge.head]] = edge_id
                tbuf[edge.head] += 1
            if edge_nr_translation:
                result.edge_ids_by_edge_nr[edge_nr] = edge_id
                edge_nr += 1

        assert result.num_edges == next_np_edge_id
        return result

    def write_dot(self, file):
        """Write a representation of the graph in dot format to `file`.

        That resulting file can be visualized with dot(1) or neato(1) form the
        graphviz package.
        """
        cdef Edge *edge
        file.write("digraph g {\n")
        for edge in self.edges[:self.size]:
            file.write("  %d -> %d;\n" % (edge.tail, edge.head))
        file.write("}\n")


cdef class gintArraySlice:
    def __len__(self):
        return self.size

    def __getitem__(self, gint key):
        cdef gint index
        if key >= 0:
            index = key
        else:
            index = self.size + key
        if index < 0 or index >= self.size:
            raise IndexError('Index out of range.')
        return self.data[index]

cdef class EdgeIterator:
    def __iter__(self):
        return self

    def __next__(self):
        if self.edge_id == self.graph.num_edges:
            raise StopIteration
        cdef gint current_edge_id = self.edge_id
        self.edge_id += 1
        while current_edge_id >= self.graph.heads_idxs[self.tail + 1]:
            self.tail += 1
        return self.tail, self.graph.heads[current_edge_id]


class DisabledFeatureError(RuntimeError):
    pass


class NodeDoesNotExistError(IndexError):
    pass


class EdgeDoesNotExistError(IndexError):
    pass


_need_twoway = 'Enable "twoway" during graph compression.'

cdef class CGraph:
    """A compressed graph which can be efficiently queried for the existence of
    edges and outgoing neighbors.

    Objects of this class do not initialize the members themselves, but expect
    that they hold usable values.  A good way to create them is by compressing
    a `Graph`.

    Iterating over a graph yields a sequence of (tail, head) pairs of all
    edges.  The number of an edge in this sequence equals its edge ID.  The
    built-in function `enumerate` can thus be used to easily iterate over all
    edges along with their edge IDs.
    """
    def __iter__(self):
        """Return an iterator over (tail, head) of all edges."""
        cdef EdgeIterator result = EdgeIterator()
        result.graph = self
        result.edge_id = 0
        result.tail = 0
        return result

    def has_dangling_edges(self):
        return not self.num_edges == self.num_px_edges == self.num_xp_edges

    cpdef gintArraySlice out_neighbors(self, gint node):
        """Return the nodes a node points to.

        Parameters
        ----------
        node : integer

        Returns
        -------
        nodes : sequence of integers

        Raises
        ------
        NodeDoesNotExistError
        """
        if node < 0 or node >= self.num_nodes:
            raise NodeDoesNotExistError()
        cdef gintArraySlice result = gintArraySlice()
        result.data = &self.heads[self.heads_idxs[node]]
        result.size = &self.heads[self.heads_idxs[node + 1]] - result.data
        return result

    def out_edge_ids(self, gint node):
        """Return the IDs of outgoing edges of node.

        Parameters
        ----------
        node : integer

        Returns
        -------
        edge_ids : sequence of integers

        Raises
        ------
        NodeDoesNotExistError
        """
        if node < 0 or node >= self.num_nodes:
            raise NodeDoesNotExistError()
        return iter(range(self.heads_idxs[node], self.heads_idxs[node + 1]))

    def in_neighbors(self, gint node):
        """Return the nodes which point to a node.

        Parameters
        ----------
        node : integer

        Returns
        -------
        nodes : sequence of integers

        Raises
        ------
        NodeDoesNotExistError
        DisabledFeatureError
            If the graph is not two-way compressed.
        """
        if not self.twoway:
            raise DisabledFeatureError(_need_twoway)
        if node < 0 or node >= self.num_nodes:
            raise NodeDoesNotExistError()
        cdef gintArraySlice result = gintArraySlice()
        result.data = &self.tails[self.tails_idxs[node]]
        result.size = &self.tails[self.tails_idxs[node + 1]] - result.data
        return result

    def in_edge_ids(self, gint node):
        """Return the IDs of incoming edges of a node.

        Parameters
        ----------
        node : integer

        Returns
        -------
        edge_ids : sequence of integers

        Raises
        ------
        NodeDoesNotExistError
        DisabledFeatureError
            If the graph is not two-way compressed.
        """
        if not self.twoway:
            raise DisabledFeatureError(_need_twoway)
        if node < 0 or node >= self.num_nodes:
            raise NodeDoesNotExistError()
        cdef gintArraySlice result = gintArraySlice()
        result.data = &self.edge_ids[self.tails_idxs[node]]
        result.size = &self.edge_ids[self.tails_idxs[node + 1]] - result.data
        return result

    def has_edge(self, gint tail, gint head):
        """Does the graph contain the edge (tail, head)?

        Parameters
        ----------
        tail : integer
        head : integer

        Returns
        -------
        had_edge : boolean

        Raises
        ------
        NodeDoesNotExistError
        EdgeDoesNotExistError
        DisabledFeatureError
            If `tail` is negative and the graph is not two-way compressed.
        """
        cdef gint h, t
        if tail >= self.num_nodes or head >= self.num_nodes:
            raise NodeDoesNotExistError()
        if tail >= 0:
            for h in self.heads[self.heads_idxs[tail]
                                : self.heads_idxs[tail + 1]]:
                if h == head: return True
        else:
            if not self.twoway:
                raise DisabledFeatureError(_need_twoway)
            if head < 0:
                raise EdgeDoesNotExistError()
            for t in self.tails[self.tails_idxs[head]
                                : self.tails_idxs[head + 1]]:
                if t == tail: return True
        return False

    def edge_id(self, gint edge_nr):
        """Return the edge ID of an edge given its sequential number.

        Parameters
        ----------
        edge_nr : integer

        Returns
        -------
        edge_id : integer

        Raises
        ------
        DisabledFeatureError
            If `edge_nr_translation` was not enabled during graph compression.
        EdgeDoesNotExistError
        """
        if not self.edge_ids_by_edge_nr:
            raise DisabledFeatureError(
                'Enable "edge_nr_translation" during graph compression.')
        if edge_nr < 0 or edge_nr >= self.edge_nr_end:
            raise EdgeDoesNotExistError()
        result = self.edge_ids_by_edge_nr[edge_nr]
        if result < 0:
            raise EdgeDoesNotExistError()
        return result

    def first_edge_id(self, gint tail, gint head):
        """Return the edge ID of the first edge (tail, head).

        Parameters
        ----------
        tail : integer
        head : integer

        Returns
        -------
        edge_id : integer

        Raises
        ------
        NodeDoesNotExist
        EdgeDoesNotExistError
        DisabledFeatureError
            If `tail` is negative and the graph is not two-way compressed.

        Notes
        -----
        This method is useful for graphs where each edge occurs only once.
        """
        if tail >= self.num_nodes or head >= self.num_nodes:
            raise NodeDoesNotExistError()
        if tail >= 0:
            for head_index in range(self.heads_idxs[tail],
                                    self.heads_idxs[tail + 1]):
                if self.heads[head_index] == head:
                    return head_index
        else:
            if not self.twoway:
                raise DisabledFeatureError(_need_twoway)
            for tail_index in range(self.tails_idxs[head],
                                    self.tails_idxs[head + 1]):
                if self.tails[tail_index] == tail:
                    return self.edge_ids[tail_index]
        raise EdgeDoesNotExistError()

    def all_edge_ids(self, gint tail, gint head):
        """Return an iterator over all edge IDs of edges with a given tail and
        head.

        Parameters
        ----------
        tail : integer
        head : integer

        Returns
        -------
        edge_id : integer

        Raises
        ------
        NodeDoesNotExist
        EdgeDoesNotExistError
        DisabledFeatureError
            If `tail` is negative and the graph is not two-way compressed.
        """
        if tail >= self.num_nodes or head >= self.num_nodes:
            raise NodeDoesNotExistError()
        result = []
        if tail >= 0:
            for head_index in range(self.heads_idxs[tail],
                                    self.heads_idxs[tail + 1]):
                if self.heads[head_index] == head:
                    result.append(head_index)
        else:
            if not self.twoway:
                raise DisabledFeatureError(_need_twoway)
            for tail_index in range(self.tails_idxs[head],
                                    self.tails_idxs[head + 1]):
                if self.tails[tail_index] == tail:
                    result.append(self.edge_ids[tail_index])
        return result

    # TODO: optimize this for the case of twofold graphs and low degree.
    def tail(self, gint edge_id):
        """Return the tail of an edge, given its edge ID.

        Parameters
        ----------
        edge_id : integer

        Returns
        -------
        tail : integer
            If the edge exists and is positive.
        None
            If the tail is negative.

        Raises
        ------
        EdgeDoesNotExistError

        Notes
        -----
        The average performance of this method is O(log num_nodes) for
        non-negative tails and O(1) for negative ones.
        """
        if edge_id < 0 or edge_id >= self.num_edges:
            raise EdgeDoesNotExistError
        if edge_id >= self.num_px_edges:
            assert self.twoway
            return None
        cdef gint lower = 0, upper = self.num_nodes, tail = 0
        while upper - lower > 1:
            tail = (upper + lower) // 2
            if edge_id == self.heads_idxs[tail]:
                return tail
            if edge_id < self.heads_idxs[tail]:
                upper = tail
            else:
                lower = tail
        return lower

    def head(self, gint edge_id):
        """Return the head of an edge, given its edge ID.

        Parameters
        ----------
        edge_id : integer

        Raises
        ------
        EdgeDoesNotExistError

        Notes
        -----
        This method executes in constant time.  It works for all edge IDs,
        returning both positive and negative heads.
        """
        if edge_id < 0 or edge_id >= self.num_edges:
            raise EdgeDoesNotExistError()
        return self.heads[edge_id]

    def write_dot(self, file):
        """Write a representation of the graph in dot format to `file`.

        Parameters
        ----------
        file : file-like object

        Notes
        -----
        That resulting file can be visualized with dot(1) or neato(1) form the
        `graphviz <http://graphviz.org/>`_ package.
        """
        cdef gint tail
        file.write("digraph g {\n")
        for tail in range(self.num_nodes):
            for head in self.heads[self.heads_idxs[tail]
                                   : self.heads_idxs[tail + 1]]:
                file.write("  %d -> %d;\n" % (tail, head))
        file.write("}\n")


cdef class CGraph_malloc(CGraph):
    """A CGraph which allocates and frees its own memory."""

    def __init__(self, twoway, edge_nr_translation, num_nodes,
                  num_pp_edges, num_pn_edges, num_np_edges):
        self.twoway = twoway
        self.edge_nr_translation = edge_nr_translation
        self.num_nodes = num_nodes
        self.num_px_edges = num_pp_edges + num_pn_edges
        self.edge_nr_end = num_pp_edges + num_pn_edges + num_np_edges
        self._heads_idxs = array.array('i', ())
        array.resize(self._heads_idxs, (num_nodes + 1))
        self.heads_idxs = <gint*>self._heads_idxs.data.as_ints
        if self.twoway:
            # The graph is two-way. n->p edges will exist in the compressed
            # graph.
            self.num_xp_edges = num_pp_edges + num_np_edges
            self.num_edges = self.edge_nr_end
            self._tails_idxs = array.array('i', ())
            array.resize(self._tails_idxs, (num_nodes + 1))
            self.tails_idxs = <gint*>self._tails_idxs.data.as_ints
            self._tails = array.array('i', ())
            array.resize(self._tails, self.num_xp_edges)
            self.tails = <gint*>self._tails.data.as_ints
            self._edge_ids = array.array('i', ())
            array.resize(self._edge_ids, self.num_xp_edges)
            self.edge_ids = <gint*>self._edge_ids.data.as_ints
        else:
            # The graph is one-way. n->p edges will be ignored.
            self.num_xp_edges = num_pp_edges
            self.num_edges = self.num_px_edges
        self._heads = array.array('i', ())
        array.resize(self._heads, self.num_edges)
        self.heads = <gint*>self._heads.data.as_ints
        if edge_nr_translation:
            self._edge_ids_by_edge_nr = array.array('i', ())
            array.resize(self._edge_ids_by_edge_nr, self.edge_nr_end)
            self.edge_ids_by_edge_nr = (<gint*>self._edge_ids_by_edge_nr
                                        .data.as_ints)
        if (not self.heads_idxs or not self.heads
            or (twoway and (not self.tails_idxs
                             or not self.tails
                             or not self.edge_ids))
            or (edge_nr_translation and not self.edge_ids_by_edge_nr)):
            raise MemoryError

    def __getstate__(self):
        twoway = self.twoway
        edge_nr_translation = self.edge_nr_translation
        num_nodes = self.num_nodes
        num_np_edges = self.edge_nr_end - self.num_px_edges
        if twoway:
            num_pp_edges = self.num_xp_edges - num_np_edges
        else:
            num_pp_edges = self.num_xp_edges
        num_pn_edges = self.num_px_edges - num_pp_edges
        init_args = (twoway, edge_nr_translation, num_nodes,
                num_pp_edges, num_pn_edges, num_np_edges)

        return init_args, (self._heads_idxs, self._heads, self._tails_idxs,
                           self._tails, self._edge_ids,
                           self._edge_ids_by_edge_nr)

    def __setstate__(self, state):
        init_args, arrays = state
        self.__init__(*init_args)
        array_attributes = (self._heads_idxs, self._heads, self._tails_idxs,
                            self._tails, self._edge_ids,
                            self._edge_ids_by_edge_nr)
        for attribute, value in zip(array_attributes, arrays):
            if attribute is None:
                continue
            attribute[:] = value

    # We are required to implement this as of Cython 0.26
    def __reduce__(self):
        state = init_args, _ = self.__getstate__()
        return (CGraph_malloc, init_args, state, None, None)



# Copyright 2011-2017 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

"""Low-level access to LAPACK functions. """

__all__ = ['getrf',
           'getrs',
           'gecon',
           'ggev',
           'gees',
           'trsen',
           'trevc',
           'gges',
           'tgsen',
           'tgevc',
           'prepare_for_lapack']

import numpy as np
cimport numpy as np

cimport scipy.linalg.cython_lapack as lapack

ctypedef int l_int
ctypedef bint l_logical

int_dtype = np.int32
logical_dtype = np.int32

ctypedef float complex float_complex
ctypedef double complex double_complex

ctypedef fused scalar:
    float
    double
    float complex
    double complex

ctypedef fused single_precision:
    float
    float complex

ctypedef fused double_precision:
    double
    double complex

ctypedef fused cmplx:
    float complex
    double complex

ctypedef fused floating:
    float
    double

# exceptions

class LinAlgError(RuntimeError):
    pass


# some helper functions
def filter_args(select, args):
    return tuple([arg for sel, arg in zip(select, args) if sel])

def assert_fortran_mat(*mats):
    # This is a workaround for a bug in NumPy version < 2.0,
    # where 1x1 matrices do not have the F_Contiguous flag set correctly.
    for mat in mats:
        if (mat is not None and (mat.shape[0] > 1 or mat.shape[1] > 1) and
            not mat.flags["F_CONTIGUOUS"]):
            raise ValueError("Input matrix must be Fortran contiguous")


cdef np.ndarray maybe_complex(scalar selector,
                              np.ndarray real, np.ndarray imag):
    cdef np.ndarray r
    r = real
    if scalar in floating:
        if imag.nonzero()[0].size:
            r = real + 1j * imag
    return r


cdef l_int lwork_from_qwork(scalar qwork):
    if scalar in floating:
        return <l_int>qwork
    else:
        return <l_int>qwork.real


def getrf(np.ndarray[scalar, ndim=2] A):
    cdef l_int M, N, info
    cdef np.ndarray[l_int] ipiv

    assert_fortran_mat(A)

    M = A.shape[0]
    N = A.shape[1]
    ipiv = np.empty(min(M,N), dtype = int_dtype)

    if scalar is float:
        lapack.sgetrf(&M, &N, <float *>A.data, &M,
                      <l_int *>ipiv.data, &info)
    elif scalar is double:
        lapack.dgetrf(&M, &N, <double *>A.data, &M,
                      <l_int *>ipiv.data, &info)
    elif scalar is float_complex:
        lapack.cgetrf(&M, &N, <float complex *>A.data, &M,
                      <l_int *>ipiv.data, &info)
    elif scalar is double_complex:
        lapack.zgetrf(&M, &N, <double complex *>A.data, &M,
                      <l_int *>ipiv.data, &info)

    assert info >= 0, "Argument error in getrf"

    return (A, ipiv, info > 0 or M != N)


def getrs(np.ndarray[scalar, ndim=2] LU, np.ndarray[l_int] IPIV,
          np.ndarray B):
    cdef l_int N, NRHS, info

    assert_fortran_mat(LU)

    # Consistency checks for LU and B

    if B.descr.type_num != LU.descr.type_num:
        raise TypeError('B must have same dtype as LU')

    # Workaround for 1x1-Fortran bug in NumPy < v2.0
    if ((B.ndim == 2 and (B.shape[0] > 1 or B.shape[1] > 1) and
         not B.flags["F_CONTIGUOUS"])):
        raise ValueError("B must be Fortran ordered")

    if B.ndim > 2:
        raise ValueError("B must be a vector or matrix")

    if LU.shape[0] != B.shape[0]:
        raise ValueError('LU and B have incompatible shapes')

    N = LU.shape[0]

    if B.ndim == 1:
        NRHS = 1
    elif B.ndim == 2:
        NRHS = B.shape[1]

    if scalar is float:
        lapack.sgetrs("N", &N, &NRHS, <float *>LU.data, &N,
                      <l_int *>IPIV.data, <float *>B.data, &N,
                      &info)
    elif scalar is double:
        lapack.dgetrs("N", &N, &NRHS, <double *>LU.data, &N,
                      <l_int *>IPIV.data, <double *>B.data, &N,
                      &info)
    elif scalar is float_complex:
        lapack.cgetrs("N", &N, &NRHS, <float complex *>LU.data, &N,
                      <l_int *>IPIV.data, <float complex *>B.data, &N,
                      &info)
    elif scalar is double_complex:
        lapack.zgetrs("N", &N, &NRHS, <double complex *>LU.data, &N,
                      <l_int *>IPIV.data, <double complex *>B.data, &N,
                      &info)

    assert info == 0, "Argument error in getrs"

    return B


def gecon(np.ndarray[scalar, ndim=2] LU, double normA, char *norm = b"1"):
    cdef l_int N, info
    cdef float srcond, snormA
    cdef double drcond

    # Parameter checks

    assert_fortran_mat(LU)
    if norm[0] != b"1" and norm[0] != b"I":
        raise ValueError("'norm' must be either '1' or 'I'")
    if scalar in single_precision:
        snormA = normA

    # Allocate workspaces

    N = LU.shape[0]

    cdef np.ndarray[l_int] iwork
    if scalar in floating:
        iwork = np.empty(N, dtype=int_dtype)

    cdef np.ndarray[scalar] work
    if scalar in floating:
        work = np.empty(4 * N, dtype=LU.dtype)
    else:
        work = np.empty(2 * N, dtype=LU.dtype)

    cdef np.ndarray rwork
    if scalar is float_complex:
        rwork = np.empty(2 * N, dtype=np.float32)
    elif scalar is double_complex:
        rwork = np.empty(2 * N, dtype=np.float64)

    # The actual calculation

    if scalar is float:
        lapack.sgecon(norm, &N, <float *>LU.data, &N, &snormA,
                      &srcond, <float *>work.data,
                      <l_int *>iwork.data, &info)
    elif scalar is double:
        lapack.dgecon(norm, &N, <double *>LU.data, &N, &normA,
                      &drcond, <double *>work.data,
                      <l_int *>iwork.data, &info)
    elif scalar is float_complex:
        lapack.cgecon(norm, &N, <float complex *>LU.data, &N, &snormA,
                      &srcond, <float complex *>work.data,
                      <float *>rwork.data, &info)
    elif scalar is double_complex:
        lapack.zgecon(norm, &N, <double complex *>LU.data, &N, &normA,
                      &drcond, <double complex *>work.data,
                      <double *>rwork.data, &info)

    assert info == 0, "Argument error in gecon"

    if scalar in single_precision:
        return srcond
    else:
        return drcond


# Helper function for xGGEV
def ggev_postprocess(dtype, alphar, alphai, vl_r=None, vr_r=None):
    # depending on whether the eigenvalues are purely real or complex,
    # some post-processing of the eigenvalues and -vectors is necessary

    indx = (alphai > 0.0).nonzero()[0]

    if indx.size:
        alpha = alphar + 1j * alphai

        if vl_r is not None:
            vl = np.array(vl_r, dtype = dtype)
            for i in indx:
                vl.imag[:, i] = vl_r[:,i+1]
                vl[:, i+1] = np.conj(vl[:, i])
        else:
            vl = None

        if vr_r is not None:
            vr = np.array(vr_r, dtype = dtype)
            for i in indx:
                vr.imag[:, i] = vr_r[:,i+1]
                vr[:, i+1] = np.conj(vr[:, i])
        else:
            vr = None
    else:
        alpha = alphar
        vl = vl_r
        vr = vr_r

    return (alpha, vl, vr)


def ggev(np.ndarray[scalar, ndim=2] A, np.ndarray[scalar, ndim=2] B,
         left=False, right=True):
    cdef l_int N, info, lwork

    # Parameter checks

    assert_fortran_mat(A, B)

    if A.ndim != 2 or A.ndim != 2:
        raise ValueError("gen_eig requires both a and be to be matrices")

    if A.shape[0] != A.shape[1]:
        raise ValueError("gen_eig requires square matrix input")

    if A.shape[0] != B.shape[0] or A.shape[1] != B.shape[1]:
        raise ValueError("A and B do not have the same shape")

    # Allocate workspaces

    N = A.shape[0]

    cdef np.ndarray[scalar] alphar, alphai
    if scalar in cmplx:
        alphar = np.empty(N, dtype=A.dtype)
        alphai = None
    else:
        alphar = np.empty(N, dtype=A.dtype)
        alphai = np.empty(N, dtype=A.dtype)

    cdef np.ndarray[scalar] beta = np.empty(N, dtype=A.dtype)

    cdef np.ndarray rwork = None
    if scalar is float_complex:
        rwork = np.empty(8 * N, dtype=np.float32)
    elif scalar is double_complex:
        rwork = np.empty(8 * N, dtype=np.float64)

    cdef np.ndarray vl
    cdef scalar *vl_ptr
    cdef char *jobvl
    if left:
        vl = np.empty((N,N), dtype=A.dtype, order='F')
        vl_ptr = <scalar *>vl.data
        jobvl = "V"
    else:
        vl = None
        vl_ptr = NULL
        jobvl = "N"

    cdef np.ndarray vr
    cdef scalar *vr_ptr
    cdef char *jobvr
    if right:
        vr = np.empty((N,N), dtype=A.dtype, order='F')
        vr_ptr = <scalar *>vr.data
        jobvr = "V"
    else:
        vr = None
        vr_ptr = NULL
        jobvr = "N"

    # Workspace query
    # Xggev expects &qwork as a <scalar *> (even though it's an integer)
    lwork = -1
    cdef scalar qwork

    if scalar is float:
        lapack.sggev(jobvl, jobvr, &N, <float *>A.data, &N,
                     <float *>B.data, &N,
                     <float *>alphar.data, <float *> alphai.data,
                     <float *>beta.data,
                     vl_ptr, &N, vr_ptr, &N,
                     &qwork, &lwork, &info)
    elif scalar is double:
        lapack.dggev(jobvl, jobvr, &N, <double *>A.data, &N,
                     <double *>B.data, &N,
                     <double *>alphar.data, <double *> alphai.data,
                     <double *>beta.data,
                     vl_ptr, &N, vr_ptr, &N,
                     &qwork, &lwork, &info)
    elif scalar is float_complex:
        lapack.cggev(jobvl, jobvr, &N, <float complex *>A.data, &N,
                     <float complex *>B.data, &N,
                     <float complex *>alphar.data, <float complex *>beta.data,
                     vl_ptr, &N, vr_ptr, &N,
                     &qwork, &lwork,
                     <float *>rwork.data, &info)
    elif scalar is double_complex:
        lapack.zggev(jobvl, jobvr, &N, <double complex *>A.data, &N,
                     <double complex *>B.data, &N,
                     <double complex *>alphar.data, <double complex *>beta.data,
                     vl_ptr, &N, vr_ptr, &N,
                     &qwork, &lwork,
                     <double *>rwork.data, &info)

    assert info == 0, "Argument error in ggev"

    lwork = lwork_from_qwork(qwork)
    cdef np.ndarray[scalar] work = np.empty(lwork, dtype=A.dtype)

    # The actual calculation

    if scalar is float:
        lapack.sggev(jobvl, jobvr, &N, <float *>A.data, &N,
                     <float *>B.data, &N,
                     <float *>alphar.data, <float *> alphai.data,
                     <float *>beta.data,
                     vl_ptr, &N, vr_ptr, &N,
                     <float *>work.data, &lwork, &info)
    elif scalar is double:
        lapack.dggev(jobvl, jobvr, &N, <double *>A.data, &N,
                     <double *>B.data, &N,
                     <double *>alphar.data, <double *> alphai.data,
                     <double *>beta.data,
                     vl_ptr, &N, vr_ptr, &N,
                     <double *>work.data, &lwork, &info)
    elif scalar is float_complex:
        lapack.cggev(jobvl, jobvr, &N, <float complex *>A.data, &N,
                     <float complex *>B.data, &N,
                     <float complex *>alphar.data, <float complex *>beta.data,
                     vl_ptr, &N, vr_ptr, &N,
                     <float complex *>work.data, &lwork,
                     <float *>rwork.data, &info)
    elif scalar is double_complex:
        lapack.zggev(jobvl, jobvr, &N, <double complex *>A.data, &N,
                     <double complex *>B.data, &N,
                     <double complex *>alphar.data, <double complex *>beta.data,
                     vl_ptr, &N, vr_ptr, &N,
                     <double complex *>work.data, &lwork,
                     <double *>rwork.data, &info)

    if info > 0:
        raise LinAlgError("QZ iteration failed to converge in sggev")

    assert info == 0, "Argument error in ggev"

    if scalar is float:
        post_dtype = np.complex64
    elif scalar is double:
        post_dtype = np.complex128

    cdef np.ndarray alpha
    alpha = alphar
    if scalar in floating:
        alpha, vl, vr = ggev_postprocess(post_dtype, alphar, alphai, vl, vr)

    return filter_args((True, True, left, right), (alpha, beta, vl, vr))


def gees(np.ndarray[scalar, ndim=2] A, calc_q=True, calc_ev=True):
    cdef l_int N, lwork, sdim, info

    assert_fortran_mat(A)

    if A.ndim != 2:
        raise ValueError("Expect matrix as input")

    if A.shape[0] != A.shape[1]:
        raise ValueError("Expect square matrix")

    # Allocate workspaces

    N = A.shape[0]

    cdef np.ndarray[scalar] wr, wi
    if scalar in cmplx:
        wr = np.empty(N, dtype=A.dtype)
        wi = None
    else:
        wr = np.empty(N, dtype=A.dtype)
        wi = np.empty(N, dtype=A.dtype)

    cdef np.ndarray rwork
    if scalar is float_complex:
        rwork = np.empty(N, dtype=np.float32)
    elif scalar is double_complex:
        rwork = np.empty(N, dtype=np.float64)

    cdef char *jobvs
    cdef scalar *vs_ptr
    cdef np.ndarray[scalar, ndim=2] vs
    if calc_q:
        vs = np.empty((N,N), dtype=A.dtype, order='F')
        vs_ptr = <scalar *>vs.data
        jobvs = "V"
    else:
        vs = None
        vs_ptr = NULL
        jobvs = "N"

    # Workspace query
    # Xgees expects &qwork as a <scalar *> (even though it's an integer)
    lwork = -1
    cdef scalar qwork

    if scalar is float:
        lapack.sgees(jobvs, "N", NULL, &N, <float *>A.data, &N,
                     &sdim, <float *>wr.data, <float *>wi.data, vs_ptr, &N,
                     &qwork, &lwork, NULL, &info)
    elif scalar is double:
        lapack.dgees(jobvs, "N", NULL, &N, <double *>A.data, &N,
                     &sdim, <double *>wr.data, <double *>wi.data, vs_ptr, &N,
                     &qwork, &lwork, NULL, &info)
    elif scalar is float_complex:
        lapack.cgees(jobvs, "N", NULL, &N, <float complex *>A.data, &N,
                     &sdim, <float complex *>wr.data, vs_ptr, &N,
                     &qwork, &lwork, <float *>rwork.data, NULL, &info)
    elif scalar is double_complex:
        lapack.zgees(jobvs, "N", NULL, &N, <double complex *>A.data, &N,
                     &sdim, <double complex *>wr.data, vs_ptr, &N,
                     &qwork, &lwork, <double *>rwork.data, NULL, &info)

    assert info == 0, "Argument error in sgees"

    lwork = lwork_from_qwork(qwork)
    cdef np.ndarray[scalar] work = np.empty(lwork, dtype=A.dtype)

    # The actual calculation

    if scalar is float:
        lapack.sgees(jobvs, "N", NULL, &N, <float *>A.data, &N,
                     &sdim, <float *>wr.data, <float *>wi.data, vs_ptr, &N,
                     <float *>work.data, &lwork, NULL, &info)
    elif scalar is double:
        lapack.dgees(jobvs, "N", NULL, &N, <double *>A.data, &N,
                     &sdim, <double *>wr.data, <double *>wi.data, vs_ptr, &N,
                     <double *>work.data, &lwork, NULL, &info)
    elif scalar is float_complex:
        lapack.cgees(jobvs, "N", NULL, &N, <float complex *>A.data, &N,
                     &sdim, <float complex *>wr.data, vs_ptr, &N,
                     <float complex *>work.data, &lwork,
                     <float *>rwork.data, NULL, &info)
    elif scalar is double_complex:
        lapack.zgees(jobvs, "N", NULL, &N, <double complex *>A.data, &N,
                     &sdim, <double complex *>wr.data, vs_ptr, &N,
                     <double complex *>work.data, &lwork,
                     <double *>rwork.data, NULL, &info)

    if info > 0:
        raise LinAlgError("QR iteration failed to converge in gees")

    assert info == 0, "Argument error in gees"

    # Real inputs possibly produce complex output
    cdef np.ndarray w = maybe_complex[scalar](0, wr, wi)

    return filter_args((True, calc_q, calc_ev), (A, vs, w))


def trsen(np.ndarray[l_logical] select,
          np.ndarray[scalar, ndim=2] T,
          np.ndarray[scalar, ndim=2] Q,
          calc_ev=True):
    cdef l_int N, M, lwork, liwork, qiwork, info

    assert_fortran_mat(T, Q)

    # Allocate workspaces

    N = T.shape[0]

    cdef np.ndarray[scalar] wr, wi
    if scalar in cmplx:
        wr = np.empty(N, dtype=T.dtype)
        wi = None
    else:
        wr = np.empty(N, dtype=T.dtype)
        wi = np.empty(N, dtype=T.dtype)

    cdef char *compq
    cdef scalar *q_ptr
    if Q is not None:
        compq = "V"
        q_ptr = <scalar *>Q.data
    else:
        compq = "N"
        q_ptr = NULL

    # Workspace query
    # Xtrsen expects &qwork as a <scalar *> (even though it's an integer)
    cdef scalar qwork
    lwork = liwork = -1

    if scalar is float:
        lapack.strsen("N", compq, <l_logical *>select.data,
                      &N, <float *>T.data, &N, q_ptr, &N,
                      <float *>wr.data, <float *>wi.data, &M, NULL, NULL,
                      &qwork, &lwork, &qiwork, &liwork, &info)
    elif scalar is double:
        lapack.dtrsen("N", compq, <l_logical *>select.data,
                      &N, <double *>T.data, &N, q_ptr, &N,
                      <double *>wr.data, <double *>wi.data, &M, NULL, NULL,
                      &qwork, &lwork, &qiwork, &liwork, &info)
    elif scalar is float_complex:
        lapack.ctrsen("N", compq, <l_logical *>select.data,
                      &N, <float complex *>T.data, &N, q_ptr, &N,
                      <float complex *>wr.data, &M, NULL, NULL,
                      &qwork, &lwork, &info)
    elif scalar is double_complex:
        lapack.ztrsen("N", compq, <l_logical *>select.data,
                      &N, <double complex *>T.data, &N, q_ptr, &N,
                      <double complex *>wr.data, &M, NULL, NULL,
                      &qwork, &lwork, &info)

    assert info == 0, "Argument error in trsen"

    lwork = lwork_from_qwork(qwork)
    cdef np.ndarray[scalar] work = np.empty(lwork, dtype=T.dtype)

    cdef np.ndarray[l_int] iwork = None
    if scalar in floating:
        liwork = qiwork
        iwork = np.empty(liwork, dtype=int_dtype)

    # Tha actual calculation

    if scalar is float:
        lapack.strsen("N", compq, <l_logical *>select.data,
                      &N, <float *>T.data, &N, q_ptr, &N,
                      <float *>wr.data, <float *>wi.data, &M, NULL, NULL,
                      <float *>work.data, &lwork,
                      <l_int *>iwork.data, &liwork, &info)
    elif scalar is double:
        lapack.dtrsen("N", compq, <l_logical *>select.data,
                      &N, <double *>T.data, &N, q_ptr, &N,
                      <double *>wr.data, <double *>wi.data, &M, NULL, NULL,
                      <double *>work.data, &lwork,
                      <l_int *>iwork.data, &liwork, &info)
    elif scalar is float_complex:
        lapack.ctrsen("N", compq, <l_logical *>select.data,
                      &N, <float complex *>T.data, &N, q_ptr, &N,
                      <float complex *>wr.data, &M, NULL, NULL,
                      <float complex *>work.data, &lwork, &info)
    elif scalar is double_complex:
        lapack.ztrsen("N", compq, <l_logical *>select.data,
                      &N, <double complex *>T.data, &N, q_ptr, &N,
                      <double complex *>wr.data, &M, NULL, NULL,
                      <double complex *>work.data, &lwork, &info)

    if info > 0:
        raise LinAlgError("Reordering failed; problem is very ill-conditioned")

    assert info == 0, "Argument error in trsen"

    # Real inputs possibly produce complex output
    cdef np.ndarray w = maybe_complex[scalar](0, wr, wi)

    return filter_args((True, Q is not None, calc_ev), (T, Q, w))


# Helper function for xTREVC and xTGEVC
def txevc_postprocess(dtype, T, vreal, np.ndarray[l_logical] select):
    cdef int N, M, i, m, indx

    N = T.shape[0]
    if select is None:
        select = np.ones(N, dtype = logical_dtype)
    selindx = select.nonzero()[0]
    M = selindx.size

    v = np.empty((N, M), dtype = dtype, order='F')

    indx = 0
    for m in range(M):
        k = selindx[m]

        if k < N-1 and T[k+1,k]:
            # we have the situation of a 2x2 block, and
            # the eigenvalue with the positive imaginary part desired
            v[:, m] = vreal[:, indx] + 1j * vreal[:, indx + 1]

            # Check if the eigenvalue with negative real part is also
            # selected, if it is, we need the same entries in vr
            if not select[k+1]:
                indx += 2
        elif k > 0 and T[k,k-1]:
            # we have the situation of a 2x2 block, and
            # the eigenvalue with the negative imaginary part desired
            v[:, m] = vreal[:, indx] - 1j * vreal[:, indx + 1]

            indx += 2
        else:
            # real eigenvalue
            v[:, m] = vreal[:, indx]

            indx += 1
    return v


def trevc(np.ndarray[scalar, ndim=2] T,
          np.ndarray[scalar, ndim=2] Q,
          np.ndarray[l_logical] select,
          left=False, right=True):
    cdef l_int N, info, M, MM
    cdef char *side
    cdef char *howmny

    # Parameter checks

    if (T.shape[0] != T.shape[1] or Q.shape[0] != Q.shape[1]
        or T.shape[0] != Q.shape[0]):
        raise ValueError("Invalid Schur decomposition as input")

    assert_fortran_mat(T, Q)

    # Workspace allocation

    N = T.shape[0]

    cdef np.ndarray[scalar] work
    if scalar in floating:
        work = np.empty(4 * N, dtype=T.dtype)
    else:
        work = np.empty(2 * N, dtype=T.dtype)

    cdef np.ndarray rwork = None
    if scalar is float_complex:
        rwork = np.empty(N, dtype=np.float32)
    elif scalar is double_complex:
        rwork = np.empty(N, dtype=np.float64)

    if left and right:
        side = "B"
    elif left:
        side = "L"
    elif right:
        side = "R"
    else:
        return

    cdef np.ndarray[l_logical] select_cpy
    cdef l_logical *select_ptr
    if select is not None:
        howmny = "S"
        MM = select.nonzero()[0].size
        # Correct for possible additional storage if a single complex
        # eigenvalue is selected.
        # For that: Figure out the positions of the 2x2 blocks.
        cmplxindx = np.diagonal(T, -1).nonzero()[0]
        for i in cmplxindx:
            if bool(select[i]) != bool(select[i+1]):
                MM += 1

        # Select is overwritten in strevc.
        select_cpy = np.array(select, dtype = logical_dtype,
                              order = 'F')
        select_ptr = <l_logical *>select_cpy.data
    else:
        MM = N
        select_ptr = NULL
        if Q is not None:
            howmny = "B"
        else:
            howmny = "A"

    cdef np.ndarray[scalar, ndim=2] vl_r = None
    cdef scalar *vl_r_ptr
    if left:
        if Q is not None and select is None:
            vl_r = np.asfortranarray(Q.copy())
        else:
            vl_r = np.empty((N, MM), dtype=T.dtype, order='F')
        vl_r_ptr = <scalar *>vl_r.data
    else:
        vl_r_ptr = NULL

    cdef np.ndarray[scalar, ndim=2]  vr_r = None
    cdef scalar *vr_r_ptr
    if right:
        if Q is not None and select is None:
            vr_r = np.asfortranarray(Q.copy())
        else:
            vr_r = np.empty((N, MM), dtype=T.dtype, order='F')
        vr_r_ptr = <scalar *>vr_r.data
    else:
        vr_r_ptr = NULL

    # The actual calculation

    if scalar is float:
        lapack.strevc(side, howmny, select_ptr,
                      &N, <float *>T.data, &N,
                      vl_r_ptr, &N, vr_r_ptr, &N, &MM, &M,
                      <float *>work.data, &info)
    elif scalar is double:
        lapack.dtrevc(side, howmny, select_ptr,
                      &N, <double *>T.data, &N,
                      vl_r_ptr, &N, vr_r_ptr, &N, &MM, &M,
                      <double *>work.data, &info)
    elif scalar is float_complex:
        lapack.ctrevc(side, howmny, select_ptr,
                      &N, <float complex *>T.data, &N,
                      vl_r_ptr, &N, vr_r_ptr, &N, &MM, &M,
                      <float complex *>work.data, <float *>rwork.data, &info)
    elif scalar is double_complex:
        lapack.ztrevc(side, howmny, select_ptr,
                      &N, <double complex *>T.data, &N,
                      vl_r_ptr, &N, vr_r_ptr, &N, &MM, &M,
                      <double complex *>work.data, <double *>rwork.data, &info)

    assert info == 0, "Argument error in trevc"
    assert MM == M, "Unexpected number of eigenvectors returned in strevc"

    if select is not None and Q is not None:
        if left:
            vl_r = np.asfortranarray(np.dot(Q, vl_r))
        if right:
            vr_r = np.asfortranarray(np.dot(Q, vr_r))

    cdef np.ndarray vl, vr
    if left:
        vl = vl_r
    if right:
        vr = vr_r
    if scalar in floating:
        # If there are complex eigenvalues, we need to postprocess the
        # eigenvectors.
        if scalar is float:
            dtype = np.complex64
        else:
            dtype = np.complex128
        if np.diagonal(T, -1).nonzero()[0].size:
            if left:
                vl = txevc_postprocess(dtype, T, vl_r, select)
            if right:
                vr = txevc_postprocess(dtype, T, vr_r, select)

    if left and right:
        return (vl, vr)
    elif left:
        return vl
    else:
        return vr


def gges(np.ndarray[scalar, ndim=2] A,
          np.ndarray[scalar, ndim=2] B,
          calc_q=True, calc_z=True, calc_ev=True):
    cdef l_int N, sdim, info

    # Check parameters

    assert_fortran_mat(A, B)

    if A.shape[0] != B.shape[1]:
        raise ValueError("Expect square matrix A")

    if A.shape[0] != B.shape[0] or A.shape[0] != B.shape[1]:
        raise ValueError("Shape of B is incompatible with matrix A")

    # Allocate workspaces

    N = A.shape[0]

    cdef np.ndarray[scalar] alphar, alphai
    if scalar in cmplx:
        alphar = np.empty(N, dtype=A.dtype)
        alphai = None
    else:
        alphar = np.empty(N, dtype=A.dtype)
        alphai = np.empty(N, dtype=A.dtype)

    cdef np.ndarray[scalar] beta = np.empty(N, dtype=A.dtype)

    cdef np.ndarray rwork = None
    if scalar is float_complex:
        rwork = np.empty(8 * N, dtype=np.float32)
    elif scalar is double_complex:
        rwork = np.empty(8 * N, dtype=np.float64)

    cdef char *jobvsl
    cdef scalar *vsl_ptr
    cdef np.ndarray[scalar, ndim=2] vsl
    if calc_q:
        vsl = np.empty((N,N), dtype=A.dtype, order='F')
        vsl_ptr = <scalar *>vsl.data
        jobvsl = "V"
    else:
        vsl = None
        vsl_ptr = NULL
        jobvsl = "N"

    cdef char *jobvsr
    cdef scalar *vsr_ptr
    cdef np.ndarray[scalar, ndim=2] vsr
    if calc_z:
        vsr = np.empty((N,N), dtype=A.dtype, order='F')
        vsr_ptr = <scalar *>vsr.data
        jobvsr = "V"
    else:
        vsr = None
        vsr_ptr = NULL
        jobvsr = "N"

    # Workspace query
    # Xgges expects &qwork as a <scalar *> (even though it's an integer)
    cdef l_int lwork = -1
    cdef scalar qwork

    if scalar is float:
        lapack.sgges(jobvsl, jobvsr, "N", NULL,
                     &N, <float *>A.data, &N,
                     <float *>B.data, &N, &sdim,
                     <float *>alphar.data, <float *>alphai.data,
                     <float *>beta.data,
                     vsl_ptr, &N, vsr_ptr, &N,
                     &qwork, &lwork, NULL, &info)
    elif scalar is double:
        lapack.dgges(jobvsl, jobvsr, "N", NULL,
                     &N, <double *>A.data, &N,
                     <double *>B.data, &N, &sdim,
                     <double *>alphar.data, <double *>alphai.data,
                     <double *>beta.data,
                     vsl_ptr, &N, vsr_ptr, &N,
                     &qwork, &lwork, NULL, &info)
    elif scalar is float_complex:
        lapack.cgges(jobvsl, jobvsr, "N", NULL,
                     &N, <float complex *>A.data, &N,
                     <float complex *>B.data, &N, &sdim,
                     <float complex *>alphar.data, <float complex *>beta.data,
                     vsl_ptr, &N, vsr_ptr, &N,
                     &qwork, &lwork, <float *>rwork.data, NULL, &info)
    elif scalar is double_complex:
        lapack.zgges(jobvsl, jobvsr, "N", NULL,
                     &N, <double complex *>A.data, &N,
                     <double complex *>B.data, &N, &sdim,
                     <double complex *>alphar.data, <double complex *>beta.data,
                     vsl_ptr, &N, vsr_ptr, &N,
                     &qwork, &lwork, <double *>rwork.data, NULL, &info)

    assert info == 0, "Argument error in gges"

    lwork = lwork_from_qwork(qwork)
    cdef np.ndarray[scalar] work = np.empty(lwork, dtype=A.dtype)

    # The actual calculation

    if scalar is float:
        lapack.sgges(jobvsl, jobvsr, "N", NULL,
                     &N, <float *>A.data, &N,
                     <float *>B.data, &N, &sdim,
                     <float *>alphar.data, <float *>alphai.data,
                     <float *>beta.data,
                     vsl_ptr, &N, vsr_ptr, &N,
                     <float *>work.data, &lwork, NULL, &info)
    elif scalar is double:
        lapack.dgges(jobvsl, jobvsr, "N", NULL,
                     &N, <double *>A.data, &N,
                     <double *>B.data, &N, &sdim,
                     <double *>alphar.data, <double *>alphai.data,
                     <double *>beta.data,
                     vsl_ptr, &N, vsr_ptr, &N,
                     <double *>work.data, &lwork, NULL, &info)
    elif scalar is float_complex:
        lapack.cgges(jobvsl, jobvsr, "N", NULL,
                     &N, <float complex *>A.data, &N,
                     <float complex *>B.data, &N, &sdim,
                     <float complex *>alphar.data, <float complex *>beta.data,
                     vsl_ptr, &N, vsr_ptr, &N,
                     <float complex *>work.data, &lwork,
                     <float *>rwork.data, NULL, &info)
    elif scalar is double_complex:
        lapack.zgges(jobvsl, jobvsr, "N", NULL,
                     &N, <double complex *>A.data, &N,
                     <double complex *>B.data, &N, &sdim,
                     <double complex *>alphar.data, <double complex *>beta.data,
                     vsl_ptr, &N, vsr_ptr, &N,
                     <double complex *>work.data, &lwork,
                     <double *>rwork.data, NULL, &info)

    if info > 0:
        raise LinAlgError("QZ iteration failed to converge in gges")

    assert info == 0, "Argument error in gges"

    # Real inputs possibly produce complex output
    cdef np.ndarray alpha = maybe_complex[scalar](0, alphar, alphai)

    return filter_args((True, True, calc_q, calc_z, calc_ev, calc_ev),
                       (A, B, vsl, vsr, alpha, beta))


def tgsen(np.ndarray[l_logical] select,
           np.ndarray[scalar, ndim=2] S,
           np.ndarray[scalar, ndim=2] T,
           np.ndarray[scalar, ndim=2] Q,
           np.ndarray[scalar, ndim=2] Z,
           calc_ev=True):
    cdef l_int ijob = 0
    cdef l_int N, M, lwork, liwork, info

    # Check parameters

    if ((S.shape[0] != S.shape[1] or T.shape[0] != T.shape[1] or
         S.shape[0] != T.shape[0]) or
        (Q is not None and (Q.shape[0] != Q.shape[1] or
                            S.shape[0] != Q.shape[0])) or
        (Z is not None and (Z.shape[0] != Z.shape[1] or
                            S.shape[0] != Z.shape[0]))):
        raise ValueError("Invalid Schur decomposition as input")

    assert_fortran_mat(S, T, Q, Z)

    # Allocate workspaces

    N = S.shape[0]

    cdef np.ndarray[scalar] alphar, alphai
    if scalar in cmplx:
        alphar = np.empty(N, dtype=S.dtype)
        alphai = None
    else:
        alphar = np.empty(N, dtype=S.dtype)
        alphai = np.empty(N, dtype=S.dtype)

    cdef np.ndarray[scalar] beta
    beta = np.empty(N, dtype=S.dtype)

    cdef l_logical wantq
    cdef scalar *q_ptr
    if Q is not None:
        wantq = 1
        q_ptr = <scalar *>Q.data
    else:
        wantq = 0
        q_ptr = NULL

    cdef l_logical wantz
    cdef scalar *z_ptr
    if Z is not None:
        wantz = 1
        z_ptr = <scalar *>Z.data
    else:
        wantz = 0
        z_ptr = NULL

    # Workspace query
    # Xtgsen expects &qwork as a <scalar *> (even though it's an integer)
    lwork = -1
    liwork = -1
    cdef scalar qwork
    cdef l_int qiwork

    if scalar is float:
        lapack.stgsen(&ijob, &wantq, &wantz, <l_logical *>select.data,
                      &N, <float *>S.data, &N,
                      <float *>T.data, &N,
                      <float *>alphar.data, <float *>alphai.data,
                      <float *>beta.data,
                      q_ptr, &N, z_ptr, &N, &M, NULL, NULL, NULL,
                      &qwork, &lwork, &qiwork, &liwork, &info)
    elif scalar is double:
        lapack.dtgsen(&ijob, &wantq, &wantz, <l_logical *>select.data,
                      &N, <double *>S.data, &N,
                      <double *>T.data, &N,
                      <double *>alphar.data, <double *>alphai.data,
                      <double *>beta.data,
                      q_ptr, &N, z_ptr, &N, &M, NULL, NULL, NULL,
                      &qwork, &lwork, &qiwork, &liwork, &info)
    elif scalar is float_complex:
        lapack.ctgsen(&ijob, &wantq, &wantz, <l_logical *>select.data,
                      &N, <float complex *>S.data, &N,
                      <float complex *>T.data, &N,
                      <float complex *>alphar.data, <float complex *>beta.data,
                      q_ptr, &N, z_ptr, &N, &M, NULL, NULL, NULL,
                      &qwork, &lwork, &qiwork, &liwork, &info)
    elif scalar is double_complex:
        lapack.ztgsen(&ijob, &wantq, &wantz, <l_logical *>select.data,
                      &N, <double complex *>S.data, &N,
                      <double complex *>T.data, &N,
                      <double complex *>alphar.data, <double complex *>beta.data,
                      q_ptr, &N, z_ptr, &N, &M, NULL, NULL, NULL,
                      &qwork, &lwork, &qiwork, &liwork, &info)

    assert info == 0, "Argument error in tgsen"

    lwork = lwork_from_qwork(qwork)
    cdef np.ndarray[scalar] work = np.empty(lwork, dtype=S.dtype)

    liwork = qiwork
    cdef np.ndarray[l_int] iwork = np.empty(liwork, dtype=int_dtype)

    # The actual calculation

    if scalar is float:
        lapack.stgsen(&ijob, &wantq, &wantz, <l_logical *>select.data,
                      &N, <float *>S.data, &N,
                      <float *>T.data, &N,
                      <float *>alphar.data, <float *>alphai.data,
                      <float *>beta.data,
                      q_ptr, &N, z_ptr, &N, &M, NULL, NULL, NULL,
                      <float *>work.data, &lwork,
                      <l_int *>iwork.data, &liwork, &info)
    elif scalar is double:
        lapack.dtgsen(&ijob, &wantq, &wantz, <l_logical *>select.data,
                      &N, <double *>S.data, &N,
                      <double *>T.data, &N,
                      <double *>alphar.data, <double *>alphai.data,
                      <double *>beta.data,
                      q_ptr, &N, z_ptr, &N, &M, NULL, NULL, NULL,
                      <double *>work.data, &lwork,
                      <l_int *>iwork.data, &liwork, &info)
    elif scalar is float_complex:
        lapack.ctgsen(&ijob, &wantq, &wantz, <l_logical *>select.data,
                      &N, <float complex *>S.data, &N,
                      <float complex *>T.data, &N,
                      <float complex *>alphar.data, <float complex *>beta.data,
                      q_ptr, &N, z_ptr, &N, &M, NULL, NULL, NULL,
                      <float complex *>work.data, &lwork,
                      <l_int *>iwork.data, &liwork, &info)
    elif scalar is double_complex:
        lapack.ztgsen(&ijob, &wantq, &wantz, <l_logical *>select.data,
                      &N, <double complex *>S.data, &N,
                      <double complex *>T.data, &N,
                      <double complex *>alphar.data, <double complex *>beta.data,
                      q_ptr, &N, z_ptr, &N, &M, NULL, NULL, NULL,
                      <double complex *>work.data, &lwork,
                      <l_int *>iwork.data, &liwork, &info)

    if info > 0:
        raise LinAlgError("Reordering failed; problem is very ill-conditioned")

    assert info == 0, "Argument error in tgsen"

    # Real inputs possibly produce complex output
    cdef np.ndarray alpha = maybe_complex[scalar](0, alphar, alphai)

    return filter_args((True, True, Q is not None, Z is not None,
                        calc_ev, calc_ev),
                       (S, T, Q, Z, alpha, beta))


def tgevc(np.ndarray[scalar, ndim=2] S,
          np.ndarray[scalar, ndim=2] T,
          np.ndarray[scalar, ndim=2] Q,
          np.ndarray[scalar, ndim=2] Z,
          np.ndarray[l_logical] select,
          left=False, right=True):
    cdef l_int N, info, M, MM

    # Check parameters

    if ((S.shape[0] != S.shape[1] or T.shape[0] != T.shape[1] or
         S.shape[0] != T.shape[0]) or
        (Q is not None and (Q.shape[0] != Q.shape[1] or
                            S.shape[0] != Q.shape[0])) or
        (Z is not None and (Z.shape[0] != Z.shape[1] or
                            S.shape[0] != Z.shape[0]))):
        raise ValueError("Invalid Schur decomposition as input")

    assert_fortran_mat(S, T, Q, Z)

    # Allocate workspaces

    N = S.shape[0]

    cdef np.ndarray[scalar] work
    if scalar in floating:
        work = np.empty(6 * N, dtype=S.dtype)
    else:
        work = np.empty(2 * N, dtype=S.dtype)

    cdef np.ndarray rwork = None
    if scalar is float_complex:
        rwork = np.empty(2 * N, dtype=np.float32)
    elif scalar is double_complex:
        rwork = np.empty(2 * N, dtype=np.float64)

    cdef char *side
    if left and right:
        side = "B"
    elif left:
        side = "L"
    elif right:
        side = "R"
    else:
        return

    cdef l_logical backtr = False

    cdef char *howmny
    cdef np.ndarray[l_logical] select_cpy = None
    cdef l_logical *select_ptr
    if select is not None:
        howmny = "S"
        MM = select.nonzero()[0].size
        # Correct for possible additional storage if a single complex
        # eigenvalue is selected.
        # For that: Figure out the positions of the 2x2 blocks.
        cmplxindx = np.diagonal(S, -1).nonzero()[0]
        for i in cmplxindx:
            if bool(select[i]) != bool(select[i+1]):
                MM += 1

        # select is overwritten in tgevc
        select_cpy = np.array(select, dtype=logical_dtype,
                              order = 'F')
        select_ptr = <l_logical *>select_cpy.data
    else:
        MM = N
        select_ptr = NULL
        if ((left and right and Q is not None and Z is not None) or
            (left and not right and Q is not None) or
            (right and not left and Z is not None)):
            howmny = "B"
            backtr = True
        else:
            howmny = "A"

    cdef np.ndarray[scalar, ndim=2] vl_r
    cdef scalar *vl_r_ptr
    if left:
        if backtr:
            vl_r = Q
        else:
            vl_r = np.empty((N, MM), dtype=S.dtype, order='F')
        vl_r_ptr = <scalar *>vl_r.data
    else:
        vl_r_ptr = NULL

    cdef np.ndarray[scalar, ndim=2] vr_r
    cdef scalar *vr_r_ptr
    if right:
        if backtr:
            vr_r = Z
        else:
            vr_r = np.empty((N, MM), dtype=S.dtype, order='F')
        vr_r_ptr = <scalar *>vr_r.data
    else:
        vr_r_ptr = NULL

    if scalar is float:
        lapack.stgevc(side, howmny, select_ptr,
                      &N, <float *>S.data, &N,
                      <float *>T.data, &N,
                      vl_r_ptr, &N, vr_r_ptr, &N, &MM, &M,
                      <float *>work.data, &info)
    elif scalar is double:
        lapack.dtgevc(side, howmny, select_ptr,
                      &N, <double *>S.data, &N,
                      <double *>T.data, &N,
                      vl_r_ptr, &N, vr_r_ptr, &N, &MM, &M,
                      <double *>work.data, &info)
    elif scalar is float_complex:
        lapack.ctgevc(side, howmny, select_ptr,
                      &N, <float complex *>S.data, &N,
                      <float complex *>T.data, &N,
                      vl_r_ptr, &N, vr_r_ptr, &N, &MM, &M,
                      <float complex *>work.data, <float *>rwork.data, &info)
    elif scalar is double_complex:
        lapack.ztgevc(side, howmny, select_ptr,
                      &N, <double complex *>S.data, &N,
                      <double complex *>T.data, &N,
                      vl_r_ptr, &N, vr_r_ptr, &N, &MM, &M,
                      <double complex *>work.data, <double *>rwork.data, &info)

    assert info == 0, "Argument error in tgevc"
    assert MM == M, "Unexpected number of eigenvectors returned in tgevc"

    if not backtr:
        if left:
            vl_r = np.asfortranarray(np.dot(Q, vl_r))
        if right:
            vr_r = np.asfortranarray(np.dot(Z, vr_r))

    # If there are complex eigenvalues, we need to postprocess the eigenvectors
    cdef np.ndarray vl, vr
    if left:
        vl = vl_r
    if right:
        vr = vr_r
    if scalar in floating:
        if scalar is float:
            dtype = np.complex64
        else:
            dtype = np.complex128
        if np.diagonal(S, -1).nonzero()[0].size:
            if left:
                vl = txevc_postprocess(dtype, S, vl_r, select)
            if right:
                vr = txevc_postprocess(dtype, S, vr_r, select)

    if left and right:
        return (vl, vr)
    elif left:
        return vl
    else:
        return vr


def prepare_for_lapack(overwrite, *args):
    """Convert arrays to Fortran format.

    This function takes a number of array objects in `args` and converts them
    to a format that can be directly passed to a Fortran function (Fortran
    contiguous NumPy array). If the arrays have different data type, they
    converted arrays are cast to a common compatible data type (one of NumPy's
    `float32`, `float64`, `complex64`, `complex128` data types).

    If `overwrite` is ``False``, an NumPy array that would already be in the
    correct format (Fortran contiguous, right data type) is neverthelessed
    copied. (Hence, overwrite = True does not imply that acting on the
    converted array in the return values will overwrite the original array in
    all cases -- it does only so if the original array was already in the
    correct format. The conversions require copying. In fact, that's the same
    behavior as in SciPy, it's just not explicitly stated there)

    If an argument is ``None``, it is just passed through and not used to
    determine the proper LAPACK type.

    Returns a list of properly converted arrays.
    """

    # Make sure we have NumPy arrays
    mats = [None]*len(args)
    for i in range(len(args)):
        if args[i] is not None:
            arr = np.asanyarray(args[i])
            if not np.issubdtype(arr.dtype, np.number):
                raise ValueError("Argument cannot be interpreted "
                                 "as a numeric array")

            mats[i] = (arr, arr is not args[i] or overwrite)
        else:
            mats[i] = (None, True)

    # First figure out common dtype
    # Note: The return type of common_type is guaranteed to be a floating point
    #       kind.
    dtype = np.common_type(*[arr for arr, ovwrt in mats if arr is not None])

    if dtype not in (np.float32, np.float64, np.complex64, np.complex128):
        raise AssertionError("Unexpected data type from common_type")

    ret = []
    for npmat, ovwrt in mats:
        # Now make sure that the array is contiguous, and copy if necessary.
        if npmat is not None:
            if npmat.ndim == 2:
                if not npmat.flags["F_CONTIGUOUS"]:
                    npmat = np.asfortranarray(npmat, dtype = dtype)
                elif npmat.dtype != dtype:
                    npmat = npmat.astype(dtype)
                elif not ovwrt:
                    # ugly here: copy makes always C-array, no way to tell it
                    # to make a Fortran array.
                    npmat = np.asfortranarray(npmat.copy())
            elif npmat.ndim == 1:
                if not npmat.flags["C_CONTIGUOUS"]:
                    npmat = np.ascontiguousarray(npmat, dtype = dtype)
                elif npmat.dtype != dtype:
                    npmat = npmat.astype(dtype)
                elif not ovwrt:
                    npmat = np.asfortranarray(npmat.copy())
            else:
                raise ValueError("Dimensionality of array is not 1 or 2")

        ret.append(npmat)

    if len(ret) == 1:
        return ret[0]
    else:
        return tuple(ret)



#!/usr/bin/env python3

# Copyright 2011-2018 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

from __future__ import print_function

import sys

import re
import os
import glob
import importlib
import subprocess
import configparser
import collections

# Until there is an alternative way to add custom build steps, request that
# setuptools' local distutils copy is used as global module "distutils".  This
# works around the future removal of distutils from the stdlib.  See
# https://github.com/pypa/setuptools/issues/2928#issuecomment-997138589.
os.environ["SETUPTOOLS_USE_DISTUTILS"] = "local"
from setuptools import setup, find_packages, Extension, Command
from distutils.errors import DistutilsError, CCompilerError
from distutils.command.build import build as build_orig
from setuptools.command.sdist import sdist as sdist_orig
from setuptools.command.build_ext import build_ext as build_ext_orig
from setuptools.command.test import test as test_orig


STATIC_VERSION_PATH = ('kwant', '_kwant_version.py')

distr_root = os.path.dirname(os.path.abspath(__file__))


def configure_extensions(exts, aliases=(), build_summary=None):
    """Modify extension configuration according to the configuration file

    `exts` must be a dict of (name, kwargs) tuples that can be used like this:
    `Extension(name, **kwargs).  This function modifies the kwargs according to
    the configuration file.

    This function modifies `sys.argv`.
    """
    global config_file, config_file_present

    #### Determine the name of the configuration file.
    config_file_option = '--configfile'
    config_file_option_present = False
    # Handle command line option
    for i, opt in enumerate(sys.argv):
        if not opt.startswith(config_file_option):
            continue
        config_file_option_present = True
        l, _, config_file = opt.partition('=')
        if l != config_file_option or not config_file:
            print('Error: Expecting {}=PATH'.format(config_file_option),
                  file=sys.stderr)
            sys.exit(1)
        sys.argv.pop(i)
        break
    else:
        config_file = 'build.conf'

    #### Read build configuration file.
    configs = configparser.ConfigParser()
    try:
        with open(config_file) as f:
            configs.read_file(f)
    except IOError:
        config_file_present = False
        if config_file_option_present:
            print("Error: '{}' option was provided, but '{}' does not exist"
                  .format(config_file_option, config_file))
            sys.exit(1)
    else:
        config_file_present = True

    #### Handle section aliases.
    for short, long in aliases:
        if short in configs:
            if long in configs:
                print('Error: both {} and {} sections present in {}.'.format(
                    short, long, config_file))
                sys.exit(1)
            configs[long] = configs[short]
            del configs[short]

    #### Apply config from file.  Use [DEFAULT] section for missing sections.
    defaultconfig = configs.defaults()
    for name, kwargs in exts.items():
        config = configs[name] if name in configs else defaultconfig
        for key, value in config.items():

            # Most, but not all, keys are lists of strings
            if key == 'language':
                pass
            elif key == 'optional':
                value = bool(int(value))
            else:
                value = value.split()

            if key == 'define_macros':
                value = [tuple(entry.split('=', maxsplit=1))
                         for entry in value]
                value = [(entry[0], None) if len(entry) == 1 else entry
                         for entry in value]

            if key in kwargs:
                msg = 'Caution: user config in file {} shadows {}.{}.'
                if build_summary is not None:
                    build_summary.append(msg.format(config_file, name, key))
            kwargs[key] = value

        kwargs.setdefault('depends', []).append(config_file)
        if config is not defaultconfig:
            del configs[name]

    unknown_sections = configs.sections()
    if unknown_sections:
        print('Error: Unknown sections in file {}: {}'.format(
            config_file, ', '.join(unknown_sections)))
        sys.exit(1)

    return exts


def check_python_version(min_version):
    installed_version = sys.version_info[:3]
    if installed_version < min_version:
        print('Error: Python {} required, but {} is installed'.format(
              '.'.join(map(str, min_version)),
              '.'.join(map(str, installed_version)))
        )
        sys.exit(1)


def check_versions():
    global version, version_is_from_git

    # Let Kwant itself determine its own version.  We cannot simply import
    # kwant, as it is not built yet.
    spec = importlib.util.spec_from_file_location('version', 'kwant/version.py')
    version_module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(version_module)

    version_module.ensure_python()
    version = version_module.version
    version_is_from_git = version_module.version_is_from_git


def init_cython():
    """Set the global variable `cythonize` (and other related globals).

    The variable `cythonize` can be in three states:

    * If Cython should be run and is ready, it contains the `cythonize()`
      function.

    * If Cython is not to be run, it contains `False`.

    * If Cython should, but cannot be run it contains `None`.  A help message
      on how to solve the problem is stored in `cython_help`.

    This function modifies `sys.argv`.
    """
    global cythonize, cython_help

    cython_option = '--cython'
    required_cython_version = (0, 24)
    try:
        sys.argv.remove(cython_option)
        cythonize = True
    except ValueError:
        cythonize = version_is_from_git

    if cythonize:
        try:
            import Cython
            from Cython.Build import cythonize
        except ImportError:
            cythonize = None
        else:
            #### Get Cython version.
            match = re.match('([0-9.]*)(.*)', Cython.__version__)
            cython_version = [int(n) for n in match.group(1).split('.')]
            # Decrease version if the version string contains a suffix.
            if match.group(2):
                while cython_version[-1] == 0:
                    cython_version.pop()
                cython_version[-1] -= 1
            cython_version = tuple(cython_version)

            if cython_version < required_cython_version:
                cythonize = None

        if cythonize is None:
            msg = ("Install Cython >= {0} or use"
                    " a source distribution (tarball) of Kwant.")
            ver = '.'.join(str(e) for e in required_cython_version)
            cython_help = msg.format(ver)
    else:
        msg = "Run setup.py with the {} option to enable Cython."
        cython_help = msg.format(cython_option)


def banner(title=''):
    starred = title.center(79, '*')
    return '\n' + starred if title else starred


class build_ext(build_ext_orig):
    def run(self):
        if not config_file_present:
            # Create an empty config file if none is present so that the
            # extensions will not be rebuilt each time.  Only depending on the
            # config file if it is present would make it impossible to detect a
            # necessary rebuild due to a deleted config file.
            with open(config_file, 'w') as f:
                f.write('# Build configuration created by setup.py '
                        '- feel free to modify.\n')

        try:
            super().run()
        except (DistutilsError, CCompilerError):
            error_msg = self.__error_msg.format(
                header=banner(' Error '), sep=banner())
            print(error_msg.format(file=config_file, summary=build_summary),
                  file=sys.stderr)
            raise
        print(banner(' Build summary '), *build_summary, sep='\n')
        print(banner())

    __error_msg = """{header}
The compilation of Kwant has failed.  Please examine the error message
above and consult the installation instructions in README.rst.
You might have to customize {{file}}.

Build configuration was:

{{summary}}
{sep}
"""


class build(build_orig):
    def run(self):
        super().run()
        write_version(os.path.join(self.build_lib, *STATIC_VERSION_PATH))


def git_lsfiles():
    if not version_is_from_git:
        return

    try:
        p = subprocess.Popen(['git', 'ls-files'], cwd=distr_root,
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except OSError:
        return

    if p.wait() != 0:
        return
    return p.communicate()[0].decode().split('\n')[:-1]


# Make the command "sdist" depend on "build".  This verifies that the
# distribution in the current state actually builds.  It also makes sure that
# the Cython-made C files will be up-to-date and included in the source.
class sdist(sdist_orig):
    sub_commands = [('build', None)] + sdist_orig.sub_commands

    def run(self):
        """Create MANIFEST.in from git if possible, otherwise check that
        MANIFEST.in is present.

        Right now (2015) generating MANIFEST.in seems to be the only way to
        include files in the source distribution that setuptools does not think
        should be there.  Setting include_package_data to True makes setuptools
        include *.pyx and other source files in the binary distribution.
        """
        manifest_in_file = 'MANIFEST.in'
        manifest = os.path.join(distr_root, manifest_in_file)
        names = git_lsfiles()
        if names is None:
            if not (os.path.isfile(manifest) and os.access(manifest, os.R_OK)):
                print("Error:", manifest_in_file,
                      "file is missing and Git is not available"
                      " to regenerate it.", file=sys.stderr)
                sys.exit(1)
        else:
            with open(manifest, 'w') as f:
                for name in names:
                    a, sep, b = name.rpartition('/')
                    if b == '.gitignore':
                        continue
                    stem, dot, extension = b.rpartition('.')
                    f.write('include {}'.format(name))
                    if extension == 'pyx':
                        f.write(''.join([' ', a, sep, stem, dot, 'c']))
                    f.write('\n')

        super().run()

        if names is None:
            msg = ("Git was not available to generate the list of files to be "
                   "included in the\nsource distribution. The old {} was used.")
            msg = msg.format(manifest_in_file)
            print(banner(' Caution '), msg, banner(), sep='\n', file=sys.stderr)

    def make_release_tree(self, base_dir, files):
        super().make_release_tree(base_dir, files)
        write_version(os.path.join(base_dir, *STATIC_VERSION_PATH))


# The following class is based on a recipe in
# http://doc.pytest.org/en/latest/goodpractices.html#manual-integration.
class test(test_orig):
    user_options = [('pytest-args=', 'a', "Arguments to pass to pytest")]

    def initialize_options(self):
        super().initialize_options()
        self.pytest_args = ''

    def run_tests(self):
        import shlex
        try:
            import pytest
        except:
            print('The Python package "pytest" is required to run tests.',
                  file=sys.stderr)
            sys.exit(1)
        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)


def write_version(fname):
    # This could be a hard link, so try to delete it first.  Is there any way
    # to do this atomically together with opening?
    try:
        os.remove(fname)
    except OSError:
        pass
    with open(fname, 'w') as f:
        f.write("# This file has been created by setup.py.\n")
        f.write("version = '{}'\n".format(version))


def long_description():
    text = []
    try:
        with open('README.rst', encoding='utf8') as f:
            for line in f:
                if line.startswith('See also in this directory:'):
                    break
                text.append(line.rstrip())
            while text[-1] == "":
                text.pop()
    except:
        return ''
    return '\n'.join(text)


def search_libs(libs):
    cmd = ['gcc']
    cmd.extend(['-l' + lib for lib in libs])
    cmd.extend(['-o/dev/null', '-xc', '-'])
    try:
        p = subprocess.Popen(cmd, stdin=subprocess.PIPE, stderr=subprocess.PIPE)
    except OSError:
        pass
    else:
        p.communicate(input=b'int main() {}\n')
        if p.wait() == 0:
            return libs


def search_mumps():
    """Return the configuration for MUMPS if it is available in a known way.

    This is known to work with the MUMPS provided by the Debian package
    libmumps-scotch-dev and the MUMPS binaries in the conda-forge channel."""
    lib_sets = [
        # Debian
        ['zmumps_scotch', 'mumps_common_scotch', 'mpiseq_scotch',
         'pord', 'gfortran'],
        # Conda (via conda-forge).
        ['zmumps_seq', 'mumps_common_seq'],
    ]
    for libs in lib_sets:
        found_libs = search_libs(libs)
        if found_libs:
            return found_libs
    return []


def configure_special_extensions(exts, build_summary):
    #### Special config for MUMPS.
    mumps = exts['kwant.linalg._mumps']
    if 'libraries' in mumps:
        build_summary.append('User-configured MUMPS')
    else:
        mumps_libs = search_mumps()
        if mumps_libs:
            mumps['libraries'] = mumps_libs
            build_summary.append('Auto-configured MUMPS')
        else:
            mumps = None
            del exts['kwant.linalg._mumps']
            build_summary.append('No MUMPS support')

    return exts


def maybe_cythonize(exts):
    """Prepare a list of `Extension` instances, ready for `setup()`.

    The argument `exts` must be a mapping of names to kwargs to be passed
    on to `Extension`.

    If Cython is to be run, create the extensions and calls `cythonize()` on
    them.  If Cython is not to be run, replace .pyx file with .c or .cpp,
    check timestamps, and create the extensions.
    """
    if cythonize:
        return cythonize([Extension(name, **kwargs)
                          for name, kwargs in exts.items()],
                         language_level=3,
                         compiler_directives={'linetrace': True})

    # Cython is not going to be run: replace pyx extension by that of
    # the shipped translated file.

    result = []
    problematic_files = []
    for name, kwargs in exts.items():
        language = kwargs.get('language')
        if language is None:
            ext = '.c'
        elif language == 'c':
            ext = '.c'
        elif language == 'c++':
            ext = '.cpp'
        else:
            print('Unknown language: {}'.format(language), file=sys.stderr)
            sys.exit(1)

        pyx_files = []
        cythonized_files = []
        sources = []
        for f in kwargs['sources']:
            if f.endswith('.pyx'):
                pyx_files.append(f)
                f = f.rstrip('.pyx') + ext
                cythonized_files.append(f)
            sources.append(f)
        kwargs['sources'] = sources

        # Complain if cythonized files are older than Cython source files.
        try:
            cythonized_oldest = min(os.stat(f).st_mtime
                                    for f in cythonized_files)
        except OSError:
            msg = "Cython-generated file {} is missing."
            print(banner(" Error "), msg.format(f), "",
                  cython_help, banner(), sep="\n", file=sys.stderr)
            sys.exit(1)

        for f in pyx_files + kwargs.get('depends', []):
            if f == config_file:
                # The config file is only a dependency for the compilation
                # of the cythonized file, not for the cythonization.
                continue
            if os.stat(f).st_mtime > cythonized_oldest:
                problematic_files.append(f)

        result.append(Extension(name, **kwargs))

    if problematic_files:
        msg = ("Some Cython source files are newer than files that have "
               "been derived from them:\n{}")
        msg = msg.format(", ".join(problematic_files))

        # Cython should be run but won't.  Signal an error if this is because
        # Cython *cannot* be run, warn otherwise.
        error = cythonize is None
        if cythonize is False:
            dontworry = ('(Do not worry about this if you are building Kwant '
                         'from unmodified sources,\n'
                         'e.g. with "pip install".)\n\n')
            msg = dontworry + msg

        print(banner(" Error " if error else " Caution "), msg, "",
              cython_help, banner(), sep="\n", file=sys.stderr)
        if error:
            sys.exit(1)

    return result


def maybe_add_numpy_include(exts):
    # Add NumPy header path to include_dirs of all the extensions.
    try:
        import numpy
    except ImportError:
        print(banner(' Caution '), 'NumPy header directory cannot be determined'
              ' ("import numpy" failed).', banner(), sep='\n', file=sys.stderr)
    else:
        numpy_include = numpy.get_include()
        for ext in exts.values():
            ext.setdefault('include_dirs', []).append(numpy_include)
    return exts


def main():
    check_python_version((3, 5))
    check_versions()

    exts = collections.OrderedDict([
        ('kwant._system',
         dict(sources=['kwant/_system.pyx'],
              include_dirs=['kwant/graph'])),
        ('kwant.operator',
         dict(sources=['kwant/operator.pyx'],
              include_dirs=['kwant/graph'])),
        ('kwant.graph.core',
         dict(sources=['kwant/graph/core.pyx'],
              depends=['kwant/graph/core.pxd', 'kwant/graph/defs.h',
                       'kwant/graph/defs.pxd'])),
        ('kwant.graph.dijkstra',
         dict(sources=['kwant/graph/dijkstra.pyx'])),
        ('kwant.linalg.lapack',
         dict(sources=['kwant/linalg/lapack.pyx'])),
        ('kwant.linalg._mumps',
         dict(sources=['kwant/linalg/_mumps.pyx'],
              depends=['kwant/linalg/cmumps.pxd']))])

    aliases = [('mumps', 'kwant.linalg._mumps')]

    init_cython()

    global build_summary
    build_summary = []
    exts = configure_extensions(exts, aliases, build_summary)
    exts = configure_special_extensions(exts, build_summary)
    exts = maybe_add_numpy_include(exts)
    exts = maybe_cythonize(exts)

    classifiers = """\
        Development Status :: 5 - Production/Stable
        Intended Audience :: Science/Research
        Intended Audience :: Developers
        Programming Language :: Python :: 3 :: Only
        Topic :: Software Development
        Topic :: Scientific/Engineering
        Operating System :: POSIX
        Operating System :: Unix
        Operating System :: MacOS :: MacOS X
        Operating System :: Microsoft :: Windows"""

    packages = find_packages('.')
    setup(name='kwant',
          version=version,
          author='C. W. Groth (CEA), M. Wimmer, '
                 'A. R. Akhmerov, X. Waintal (CEA), and others',
          author_email='authors@kwant-project.org',
          description=("Package for numerical quantum transport calculations "
                       "(Python 3 version)"),
          long_description=long_description(),
          platforms=["Unix", "Linux", "Mac OS-X", "Windows"],
          url="http://kwant-project.org/",
          license="BSD",
          packages=packages,
          package_data={p: ['*.pxd', '*.h'] for p in packages},
          cmdclass={'build': build,
                    'sdist': sdist,
                    'build_ext': build_ext,
                    'test': test},
          ext_modules=exts,
          install_requires=['numpy >= 1.11.0', 'scipy >= 0.17.0',
                            'tinyarray >= 1.2'],
          extras_require={
              # The oldest versions between: Debian stable, Ubuntu LTS
              'plotting': 'matplotlib >= 1.5.1',
              'continuum': 'sympy >= 0.7.6',
              # qsymm is only packaged on PyPI
              'qsymm': 'qsymm >= 1.2.6',
          },
          classifiers=[c.strip() for c in classifiers.split('\n')])

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
#
# Kwant documentation build configuration file, created by
# sphinx-quickstart on Tue Jan 11 09:39:28 2011.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys
import os
import string
from distutils.util import get_platform
sys.path.insert(0, "../../build/lib.{0}-{1}.{2}".format(
        get_platform(), *sys.version_info[:2]))

import kwant
import kwant.qsymm
import kwant.continuum  # sphinx gets confused with lazy loading

# -- General configuration -----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.

sys.path.insert(0, os.path.abspath('../sphinxext'))

extensions = ['sphinx.ext.autodoc', 'sphinx.ext.autosummary',
              'sphinx.ext.todo', 'sphinx.ext.mathjax', 'numpydoc',
              'kwantdoc', 'sphinx.ext.linkcode']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['../templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'Kwant'
copyright = '2011-2017, C. W. Groth (CEA), M. Wimmer, A. R. Akhmerov, X. Waintal (CEA), and others'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The full version, including alpha/beta/rc tags.
release = kwant.__version__

# The short X.Y version.
version = release[:len(release) - len(release.lstrip('012345679.'))].rstrip('.')

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# List of documents that shouldn't be included in the build.
#unused_docs = []

# List of directories, relative to source directory, that shouldn't be searched
# for source files.
exclude_trees = []

# The reST default role (used for this markup: `text`) to use for all documents.
default_role = "autolink"

# If true, '()' will be appended to :func: etc. cross-reference text.
#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []

# Do not show all class members automatically in the class documentation
numpydoc_show_class_members = False

# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
html_theme = 'kwantdoctheme'
html_theme_path = ['..']
html_theme_options = {'collapsiblesidebar': True}
html_style = 'kwant.css'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
#html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
#html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
html_use_modindex = False

# This is needed too.
html_domain_indices = False

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
#html_show_sourcelink = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# If nonempty, this is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = ''

# Output file base name for HTML help builder.
htmlhelp_basename = 'kwantdoc'


# -- Options for LaTeX output --------------------------------------------------

# http://thread.gmane.org/gmane.comp.python.sphinx.devel/4220/focus=4238
latex_elements = {'papersize': 'a4paper',
                  'release': '',
                  'releasename': '',
                  'preamble':
r"""\makeatletter
  \fancypagestyle{normal}{
    \fancyhf{}
    \fancyfoot[LE,RO]{{\py@HeaderFamily\thepage}}
    \fancyfoot[LO]{{\py@HeaderFamily\nouppercase{\rightmark}}}
    \fancyfoot[RE]{{\py@HeaderFamily\nouppercase{\leftmark}}}
    \fancyhead[LE,RO]{{\py@HeaderFamily \@title}}
    \renewcommand{\headrulewidth}{0.4pt}
    \renewcommand{\footrulewidth}{0.4pt}
  }
\makeatother

\newcommand{\bra}[1]{\left\langle#1\right|}
\newcommand{\ket}[1]{\left|#1\right>}
\newcommand{\braket}[2]{\left\langle#1|#2\right\rangle}
\newcommand{\ri}{\text{i}}
\newcommand{\rd}{\text{d}}

\usepackage{unicode-math}
"""}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
# We use "et al." as it is shorter and there's not much space left horizontally.
latex_documents = [
  ('index', 'kwant.tex', 'Kwant {0} documentation'.format(release),
   'C. W. Groth, M. Wimmer, A. R. Akhmerov, X. Waintal, et al.',
   'manual'),
]

latex_engine = 'xelatex'

latex_use_xindy = False

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#latex_use_parts = False

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# This is needed too.
latex_domain_indices = False

# -- Options for autodoc -------------------------------------------------------
# Generate stub pages for autosummary directives.
autosummary_generate = True

autoclass_content = "both"
autodoc_default_options = {
    'show-inheritance': True,
}

# -- Teach Sphinx to document bound methods like functions ---------------------
import types
from sphinx.ext import autodoc

class BoundMethodDocumenter(autodoc.FunctionDocumenter):
    objtype = "boundmethod"
    directivetype = 'function'

    @classmethod
    def can_document_member(cls, member, membername, isattr, parent):
        # Return True iff `member` is a bound method.  Taken from
        # <http://stackoverflow.com/a/1260881>.
        return (isinstance(member, types.MethodType) and
                member.__self__ is not None and
                not issubclass(member.__self__.__class__, type) and
                member.__self__.__class__ is not type)

    def format_args(self):
        args = super(BoundMethodDocumenter, self).format_args()
        left, sep, right = args.partition('self, ')
        if left.endswith('('):
            args = left + right
        return args

def setup(app):
    app.add_autodocumenter(BoundMethodDocumenter)


# By default linkcheck only tries once, but experience has shown us that this test is
# very non-specific for detecting dead links. Trying a few times should significantly
# lower the probability of false positives.
linkcheck_retries = 5
# Some websites (particularly IoP) take a long time to respond. The combination of
# this timeout and the retries should make this check sufficiently specific.
linkcheck_timeout = 10

nitpick_ignore = [('py:class', 'Warning'), ('py:class', 'Exception'),
                  ('py:class', 'object'), ('py:class', 'tuple'),
                  ('py:class', 'kwant.operator._LocalOperator'),
                  ('py:class', 'numpy.ndarray'),
                  ('py:class', 'kwant.solvers.common.BlockResult'),
                  ('py:class', 'kwant.builder._FinalizedBuilderMixin')]


# Use custom MathJax CDN, as cdn.mathjax.org will soon shut down
mathjax_path = "https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.0/MathJax.js?config=TeX-AMS-MML_HTMLorMML"


# -- Make Sphinx insert source code links --------------------------------------
def linkcode_resolve(domain, info):
    def find_source():
        # try to find the file and line number, based on code from numpy:
        # https://github.com/numpy/numpy/blob/master/doc/source/conf.py#L286
        obj = sys.modules[info['module']]
        for part in info['fullname'].split('.'):
            obj = getattr(obj, part)
        import inspect
        fn = inspect.getsourcefile(obj)
        fn = os.path.relpath(fn, start=os.path.dirname(kwant.__file__))
        source, lineno = inspect.getsourcelines(obj)
        return fn, lineno, lineno + len(source) - 1

    if domain != 'py' or not info['module']:
        return None
    try:
        filename = 'kwant/%s#L%d-%d' % find_source()
    except Exception:
        filename = info['module'].replace('.', '/') + '.py'

    # The following relies on the documented format of kwant.__version__.
    rel = release.rstrip('.dirty')
    if '.dev' in rel:
        try:
            refname = rel[rel.index('+g') + 2:]
        except ValueError:
            return
    else:
        refname = 'v' + rel

    templ = os.environ.get(
        "SOURCE_LINK_TEMPLATE",
        "https://gitlab.kwant-project.org/kwant/kwant/blob/$r/$f")
    return string.Template(templ).safe_substitute(r=refname, f=filename)



# Copyright 2011-2013 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

cimport numpy as np
import numpy as np
from . cimport cmumps
from . import cmumps
from .fortran_helpers import assert_fortran_matvec, assert_fortran_mat

int_dtype = cmumps.int_dtype

# Proxy classes for Python access to the control and info parameters of MUMPS

cdef class mumps_int_array:
    cdef cmumps.MUMPS_INT *array

    def __init__(self):
        self.array = NULL

    def __getitem__(self, key):
        return self.array[key - 1]

    def __setitem__(self, key, value):
        self.array[key - 1] = value


# workaround for the fact that cython cannot pass pointers to an __init__()
cdef make_mumps_int_array(cmumps.MUMPS_INT *array):
    wrapper = mumps_int_array()
    wrapper.array = array
    return wrapper


cdef class smumps_real_array:
    cdef cmumps.SMUMPS_REAL *array

    def __init__(self):
        self.array = NULL

    def __getitem__(self, key):
        return self.array[key - 1]

    def __setitem__(self, key, value):
        self.array[key - 1] = value


cdef make_smumps_real_array(cmumps.SMUMPS_REAL *array):
    wrapper = smumps_real_array()
    wrapper.array = array
    return wrapper


cdef class dmumps_real_array:
    cdef cmumps.DMUMPS_REAL *array

    def __init__(self):
        self.array = NULL

    def __getitem__(self, key):
        return self.array[key - 1]

    def __setitem__(self, key, value):
        self.array[key - 1] = value


cdef make_dmumps_real_array(cmumps.DMUMPS_REAL *array):
    wrapper = dmumps_real_array()
    wrapper.array = array
    return wrapper


cdef class cmumps_real_array:
    cdef cmumps.CMUMPS_REAL *array

    def __init__(self):
        self.array = NULL

    def __getitem__(self, key):
        return self.array[key - 1]

    def __setitem__(self, key, value):
        self.array[key - 1] = value


cdef make_cmumps_real_array(cmumps.CMUMPS_REAL *array):
    wrapper = cmumps_real_array()
    wrapper.array = array
    return wrapper


cdef class zmumps_real_array:
    cdef cmumps.ZMUMPS_REAL *array

    def __init__(self):
        self.array = NULL

    def __getitem__(self, key):
        return self.array[key - 1]

    def __setitem__(self, key, value):
        self.array[key - 1] = value


cdef make_zmumps_real_array(cmumps.ZMUMPS_REAL *array):
    wrapper = zmumps_real_array()
    wrapper.array = array
    return wrapper

#############################################################

cdef class zmumps:
    cdef cmumps.ZMUMPS_STRUC_C params

    cdef public mumps_int_array icntl
    cdef public zmumps_real_array cntl
    cdef public mumps_int_array info
    cdef public mumps_int_array infog
    cdef public zmumps_real_array rinfo
    cdef public zmumps_real_array rinfog

    def __init__(self, verbose=False, sym=0):
        self.params.job = -1
        self.params.sym = sym
        self.params.par = 1
        self.params.comm_fortran = -987654

        cmumps.zmumps_c(&self.params)

        self.icntl = make_mumps_int_array(self.params.icntl)
        self.cntl = make_zmumps_real_array(self.params.cntl)
        self.info = make_mumps_int_array(self.params.info)
        self.infog = make_mumps_int_array(self.params.infog)
        self.rinfo = make_zmumps_real_array(self.params.rinfo)
        self.rinfog = make_zmumps_real_array(self.params.rinfog)

        # no diagnostic output (MUMPS is very verbose normally)
        if not verbose:
            self.icntl[1] = 0
            self.icntl[3] = 0

    def __dealloc__(self):
        self.params.job = -2
        cmumps.zmumps_c(&self.params)

    def call(self):
        cmumps.zmumps_c(&self.params)

    def _set_job(self, value):
        self.params.job = value

    def _get_job(self):
        return self.params.job

    job = property(_get_job, _set_job)

    @property
    def sym(self):
        return self.params.sym

    def set_assembled_matrix(self,
                             cmumps.MUMPS_INT N,
                             np.ndarray[cmumps.MUMPS_INT, ndim=1] i,
                             np.ndarray[cmumps.MUMPS_INT, ndim=1] j,
                             np.ndarray[np.complex128_t, ndim=1] a):
        self.params.n = N
        self.params.nz = a.shape[0]
        self.params.irn = <cmumps.MUMPS_INT *>i.data
        self.params.jcn = <cmumps.MUMPS_INT *>j.data
        self.params.a = <cmumps.ZMUMPS_COMPLEX *>a.data

    def set_dense_rhs(self, np.ndarray rhs):

        assert_fortran_matvec(rhs)
        if rhs.dtype != np.complex128:
            raise ValueError("numpy array must be of dtype complex128!")

        if rhs.ndim == 1:
            self.params.nrhs = 1
        else:
            self.params.nrhs = rhs.shape[1]
        self.params.lrhs = rhs.shape[0]
        self.params.rhs = <cmumps.ZMUMPS_COMPLEX *>rhs.data

    def set_sparse_rhs(self,
                       np.ndarray[cmumps.MUMPS_INT, ndim=1] col_ptr,
                       np.ndarray[cmumps.MUMPS_INT, ndim=1] row_ind,
                       np.ndarray[np.complex128_t, ndim=1] data):

        if row_ind.shape[0] != data.shape[0]:
            raise ValueError("Number of entries in row index and value "
                             "array differ!")

        self.params.nz_rhs = data.shape[0]
        self.params.nrhs = col_ptr.shape[0] - 1
        self.params.rhs_sparse = <cmumps.ZMUMPS_COMPLEX *>data.data
        self.params.irhs_sparse = <cmumps.MUMPS_INT *>row_ind.data
        self.params.irhs_ptr = <cmumps.MUMPS_INT *>col_ptr.data

    def set_schur(self,
                  np.ndarray[np.complex128_t, ndim=2, mode='c'] schur,
                  np.ndarray[cmumps.MUMPS_INT, ndim=1] schur_vars):

        if schur.shape[0] != schur.shape[1]:
            raise ValueError("Schur matrix must be squared!")
        if schur.shape[0] != schur_vars.shape[0]:
            raise ValueError("Number of Schur variables must agree "
                             "with Schur complement size!")

        self.params.size_schur = schur.shape[0]
        self.params.schur = <cmumps.ZMUMPS_COMPLEX *>schur.data
        self.params.listvar_schur = <cmumps.MUMPS_INT *>schur_vars.data
